// -*- C++ -*-

#pragma once

#include "gpu_types.h"
#include "il_writer_utils.h"
#include "memory.h"

/**
 \addtogroup atomic atomic operations
atomic functions are used to atomically modify values in global or local memory.

 example:
 \code{.cpp}
 struct foo :
 kernel<foo>
 {
    void program(resource<int>& a, resource<pair<uint64_t, double>>& B)
    {
        atomic_add(a[0], 1);                               // increments a[0] by 1
        gpu_double x = atomic_xchg(b[12].second, 3.14);    // Writes 3.14 at B[12].second, returns in x the previous value
    }
 };
 \endcode

 \{
*/

namespace goopax {
namespace impl {
namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_atomic_func2(gpu_writer_base_impl*, gpu_writer_base_impl*, Uint) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_atomic_cmpxchg2(gpu_writer_base_impl*, gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
}
}


template<class gpu_T>
gpu_T call_atomic(gpu_T& ref, const gpu_T& value, Tuint N) noexcept
{
    gpu_T ret = value;
    check_errno(interface::goopax_E_atomic_func2(ref.impl.get(), ret.impl.get(), N));
    return ret;
}

#define MAKEFUNCT(F, gpu_T, N)                             \
    inline gpu_T atomic_##F(gpu_T& ref, const gpu_T value) \
    {                                                      \
        return call_atomic<gpu_T>(ref, value, N);          \
    }
#define MAKEFUNC(F, N)          \
    MAKEFUNCT(F, gpu_uint32, N) \
    MAKEFUNCT(F, gpu_int32, N)  \
    MAKEFUNCT(F, gpu_uint64, N) \
    MAKEFUNCT(F, gpu_int64, N)

/// atomic and
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral type.
/// \param value value to perform and operation with
/// \return previous value
MAKEFUNC(and, 1)
/// atomic or
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral type.
/// \param value value to perform or operation with
/// \return previous value
MAKEFUNC(or, 2)
/// atomic xor
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral type.
/// \param value value to perform xor operation with
/// \return previous value
MAKEFUNC(xor, 3)
/// atomic min
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral type.
/// \param value value to perform min operation with
/// \return previous value
MAKEFUNC(min, 4)
/// atomic max
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral type.
/// \param value value to perform max operation with
/// \return previous value
MAKEFUNC(max, 5)
/// atomic exchange
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral or floating point type.
/// \param value value to write
/// \return previous value
MAKEFUNC(xchg, 6)
/// atomic exchange
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral or floating point type.
/// \param value value to write
/// \return previous value
MAKEFUNCT(xchg, gpu_float, 6)
/// atomic exchange
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral or floating point type.
/// \param value value to write
/// \return previous value
MAKEFUNCT(xchg, gpu_double, 6)

#undef MAKEFUNC
#undef MAKEFUNCT

/// atomic add
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral type.
/// \param value value to add
/// \return previous value
inline gpu_uint32 atomic_add(gpu_uint32& ref, const gpu_uint32 value)
{
    return call_atomic<gpu_uint32>(ref, value, 0);
}
inline gpu_int32 atomic_add(gpu_int32& ref, const gpu_int32 value)
{
    return call_atomic<gpu_int32>(ref, value, 0);
}
inline gpu_uint64 atomic_add(gpu_uint64& ref, const gpu_uint64 value)
{
    return call_atomic<gpu_uint64>(ref, value, 0);
}
inline gpu_int64 atomic_add(gpu_int64& ref, const gpu_int64 value)
{
    return call_atomic<gpu_int64>(ref, value, 0);
}
inline gpu_float atomic_add(gpu_float& ref, const gpu_float value)
{
    return call_atomic<gpu_float>(ref, value, 0);
}
inline gpu_double atomic_add(gpu_double& ref, const gpu_double value)
{
    return call_atomic<gpu_double>(ref, value, 0);
}


/// atomic load
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral or floating point type.
/// \return value at location 'ref'
inline gpu_uint32 atomic_load(const gpu_uint32& ref)
{
    return atomic_add(const_cast<gpu_uint32&>(ref), 0);
}
/// atomic load
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral or floating point type.
/// \return value at location 'ref'
inline gpu_int32 atomic_load(const gpu_int32& ref)
{
    return atomic_add(const_cast<gpu_int32&>(ref), 0);
}
/// atomic load
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral or floating point type.
/// \return value at location 'ref'
inline gpu_uint64 atomic_load(const gpu_uint64& ref)
{
    return atomic_add(const_cast<gpu_uint64&>(ref), 0);
}
/// atomic load
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral or floating point type.
/// \return value at location 'ref'
inline gpu_int64 atomic_load(const gpu_int64& ref)
{
    return atomic_add(const_cast<gpu_int64&>(ref), 0);
}

/*
      /// atomic load
      /// \ingroup atomic
      ///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral or floating point type.
      /// \return value at location 'ref'
      gpu_float atomic_load(const gpu_float& ref)
      {
	  return atomic_add(const_cast<gpu_float&>(ref), 0);
      }
      /// atomic load
      /// \ingroup atomic
      ///  \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral or floating point type.
      /// \return value at location 'ref'
      gpu_double atomic_load(const gpu_double& ref)
      {
	  return atomic_add(const_cast<gpu_double&>(ref), 0);
      }
*/

#define MAKETYPE(gpu_T)                                                                                                                               \
    inline gpu_T atomic_sub(gpu_T& ref, gpu_T value)                                                                                                  \
    {                                                                                                                                                 \
        /** atomic sub */                                                                                                                             \
        /** \ingroup atomic */                                                                                                                        \
        /** \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral type. */ \
        /** \return previous value */                                                                                                                 \
        return atomic_add(static_cast<gpu_T&>(ref), -value);                                                                                          \
    }                                                                                                                                                 \
    inline gpu_T atomic_cmpxchg(gpu_T& ref, gpu_T cmp, gpu_T value)                                                                                   \
    {                                                                                                                                                 \
        /** atomic compare and exchange */                                                                                                            \
        /** \ingroup atomic */                                                                                                                        \
        /** If ref == cmp, write ref = value. Otherwise do nothing. */                                                                                \
        /** \param ref reference to local or global memory address. The value referenced must be of 32 or 64 bit signed or unsigned integral type. */ \
        /** \param cmp value to compare with */                                                                                                       \
        /** \param value value to write */                                                                                                            \
        /** \return previous value */                                                                                                                 \
        gpu_T ret = value;                                                                                                                            \
        check_errno(interface::goopax_E_atomic_cmpxchg2(ref.impl.get(), cmp.impl.get(), ret.impl.get()));                                             \
        return ret;                                                                                                                                   \
    }

MAKETYPE(gpu_int32)
MAKETYPE(gpu_uint32)
MAKETYPE(gpu_int64)
MAKETYPE(gpu_uint64)
#undef MAKETYPE

}
}


/// \}
