// -*- C++ -*-


#pragma once

/// \addtogroup OpenCL OpenCL interoperability
/// OpenCL interoperability functions. These can be used to use existing OpenCL kernels in conjunktion with GOOPAX.
/// \{


#include "lib/defs.h"
#if GOOPAX_HAVE_OPENCL

#include <string>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#if defined(CL_VERSION_2_1)
#define GOOPAX_OPENCL_VERSION 20100
#elif defined(CL_VERSION_2_0)
#define GOOPAX_OPENCL_VERSION 20000
#elif defined(CL_VERSION_1_2)
#define GOOPAX_OPENCL_VERSION 10200
#elif defined(CL_VERSION_1_1)
#define GOOPAX_OPENCL_VERSION 10100
#else
#error "Sorry, GOOPAX requires OpenCL version 1.1 or higher."
#endif


#include "lib/clean_environment.h"

#include "env.h"
#include "kernel.h"
#include "memory.h"


namespace goopax {
namespace impl {
namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_get_cl_platform2(cl_platform_id*, const goopax_device_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_get_cl_context2(cl_context*, const goopax_device_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_get_cl_queue2(cl_command_queue*, const goopax_device_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_get_cl_device2(cl_device_id*, const goopax_device_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_num_cl_platform_extensions2(Size_t*, const goopax_device_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) const char* goopax_get_cl_platform_extension2(const goopax_device_impl*, Size_t num) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_num_cl_device_extensions2(Size_t*, const goopax_device_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) const char* goopax_get_cl_device_extension2(const goopax_device_impl*, Size_t num) noexcept;

GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_have_cl_platform_extension(Bool*, const char*, const goopax_device_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_have_cl_device_extension(Bool*, const char*, const goopax_device_impl*) noexcept;


GOOPAX_DLL_EXPORT(3.0.0) cl_mem goopax_get_cl_mem2(const texbuf* buf) noexcept;

GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_create_device_from_cl(goopax_device_impl**, cl_command_queue, Uint, const char*, const Uint8_t*) noexcept;
GOOPAX_DLL_EXPORT(3.0.3) errno_type goopax_E_create_device_from_cl2(goopax_device_impl**, cl_command_queue, Uint, void (*)(), const char*, const Uint8_t*) noexcept;
}
}

/// \ingroup OpenCL
/// Create image buffer from existing OpenCL image.
/// \param device: goopax device that corresponds to the OpenCL device
/// \param mem OpenCL memory resource
template<Uint DIM, typename T, Bool normalized>
image_buffer<DIM, T, normalized> image_buffer<DIM, T, normalized>::create_from_cl(goopax_device device, cl_mem mem)
{
    return image_buffer(INITTEX_IMPL(),
                        if_call(interface::goopax_E_image_create_cl2, device.get_impl(), nullptr, DIM, false, mem, image_buffer::channel_type_enum, image_buffer::num_channels, normalized));
}

template<Uint DIM, typename T, Bool normalized>
image_array_buffer<DIM, T, normalized> image_array_buffer<DIM, T, normalized>::create_from_cl(goopax_device device, cl_mem mem)
{
    /// \ingroup OpenCL
    /// Create image array buffer from existing OpenCL image array.
    /// \param device: goopax device that corresponds to the OpenCL device
    /// \param mem OpenCL memory resource
    return image_array_buffer(INITTEX_IMPL(),
                              if_call(interface::goopax_E_image_create_cl2, device, nullptr, DIM, true, mem, image_array_buffer::channel_type_enum, image_array_buffer::num_channels, normalized));
}


inline cl_platform_id get_cl_platform(goopax_device device)
{
    /// \ingroup OpenCL
    /// Get OpenCL platform from goopax device
    return if_call(interface::goopax_E_get_cl_platform2, device.get_impl());
}
inline cl_context get_cl_context(const goopax_device device)
{
    /// \ingroup OpenCL
    /// Get OpenCL context from goopax device
    return if_call(interface::goopax_E_get_cl_context2, device.get_impl());
}

inline cl_command_queue get_cl_queue(const goopax_device device)
{
    /// \ingroup OpenCL
    /// Get OpenCL queue from goopax device
    return if_call(interface::goopax_E_get_cl_queue2, device.get_impl());
}
inline cl_device_id get_cl_device(const goopax_device device)
{
    /// \ingroup OpenCL
    /// Get OpenCL device from goopax device
    return if_call(interface::goopax_E_get_cl_device2, device.get_impl());
}
inline vector<string> get_cl_platform_extensions(const goopax_device device)
{
    /// \ingroup OpenCL
    /// Get OpenCL platform extensions for the given goopax device
    vector<string> ret(if_call(interface::goopax_E_num_cl_platform_extensions2, device.get_impl()));
    for (goopax::impl::Tsize_t k = 0; k < ret.size(); ++k)
    {
        ret[k] = interface::goopax_get_cl_platform_extension2(device.get_impl(), k);
    }
    return ret;
}
inline vector<string> get_cl_device_extensions(const goopax_device device)
{
    /// \ingroup OpenCL
    /// Get OpenCL device extensions for the given goopax device
    vector<string> ret(if_call(interface::goopax_E_num_cl_device_extensions2, device.get_impl()));
    for (goopax::impl::Tsize_t k = 0; k < ret.size(); ++k)
    {
        ret[k] = interface::goopax_get_cl_device_extension2(device.get_impl(), k);
    }
    return ret;
}

inline bool have_cl_platform_extension(const string& s, const goopax_device device)
{
    /// \ingroup OpenCL
    /// Test if OpenCL platform extension is available for the device.
    return if_call(interface::goopax_E_have_cl_platform_extension, s.c_str(), device.get_impl());
}
inline bool have_cl_device_extension(const string& s, const goopax_device device)
{
    /// \ingroup OpenCL
    /// Test if OpenCL device extension is available for the device.
    return if_call(interface::goopax_E_have_cl_device_extension, s.c_str(), device.get_impl());
}

template<class BUF>
inline cl_mem get_cl_mem(const BUF& buf)
{
    /// \ingroup OpenCL
    /// Returns OpenCL memory handle for the goopax buffer or image_buffer
    /// \param buf buffer, image_buffer, or image_array_buffer
    return interface::goopax_get_cl_mem2(buf.get_buf());
}

inline goopax_device get_device_from_cl_queue(cl_command_queue queue)
{
    /// \ingroup OpenCL
    /// Create goopax device from OpenCL queue
    goopax_device ret;
    ret.impl = if_call(interface::goopax_E_create_device_from_cl2, queue, Uint(GOOPAX_VERSION_ID), goopax_debugfunc, GOOPAX_CERT, sig);
    return ret;
}


#ifdef CL_HPP_
/// \ingroup OpenCL
/// \name OpenCL C++ interface functions
/// Include cl.hpp before goopax_cl.h to enable cl.hpp functionalities.
/// \{

inline cl::Context get_cl_cxx_context(goopax_device device)
{
    /// Get cl::Context for the given goopax device.
    cl::Context ret;
    ret() = get_cl_context(device);
    clRetainContext(ret());
    return ret;
}
inline cl::CommandQueue get_cl_cxx_queue(goopax_device device)
{
    /// Get cl::CommandQueue from goopax device
    cl::CommandQueue ret;
    ret() = get_cl_queue(device);
    clRetainCommandQueue(ret());
    return ret;
}
#if GOOPAX_OPENCL_VERSION < 10200
inline cl::Device get_cl_cxx_device(goopax_device device)
{
    /// Get cl::Device from goopax device
    return if_call(interface::goopax_E_get_cl_device2, device.get_impl());
}
#else
inline cl::Device get_cl_cxx_device(const goopax_device device)
{
    /// Get cl::Device from goopax device
    cl::Device ret;
    ret() = if_call(interface::goopax_E_get_cl_device2, device.get_impl());
    clRetainDevice(ret());
    return ret;
}
#endif
template<class BUF>
inline cl::Buffer get_cl_cxx_buf(const BUF& buf)
{
    /// Get cl::Buffer from goopax buffer, image_buffer or image_array_buffer
    cl::Buffer ret;
    ret() = get_cl_mem(buf);
    clRetainMemObject(ret());
    return ret;
}

/// \}
#endif


}

using impl::get_cl_context;
using impl::get_cl_device;
using impl::get_cl_device_extensions;
using impl::get_cl_mem;
using impl::get_cl_platform;
using impl::get_cl_platform_extensions;
using impl::get_cl_queue;
using impl::get_device_from_cl_queue;
using impl::have_cl_device_extension;
using impl::have_cl_platform_extension;

#ifdef CL_HPP_
using impl::get_cl_cxx_buf;
using impl::get_cl_cxx_context;
using impl::get_cl_cxx_device;
using impl::get_cl_cxx_queue;
#endif
}
#endif

/// \}
