// -*- C++ -*-
#pragma once

#include "./lib/range.h"
#include "lib/functions.h"
#include "operators.h"


/// @{
/// @addtogroup functions cpu and gpu functions
/// \brief CPU functions
/// In the namespace goopax::extra_operators additional function definitions for CPU data types are provided that otherwise are only provided for gpu data types.

namespace goopax {
namespace impl {
#if !defined(__GNUC__) || (__GNUC__ == 7 && __GNUC_MINOR__ < 2)
// Using our own functions

inline Uint __builtin_popcount(Uint i)
{
    Uint ret = 0;
    while (i != 0)
    {
        ret += (i & 1);
        i >>= 1;
    }
    return ret;
}
inline Uint __builtin_clz(Uint i)
{
    Uint ret = 0;
    while ((i & 0x80000000) == 0 && ret < 32)
    {
        ++ret;
        i <<= 1;
    }
    return ret;
}
#endif

template<typename T, typename = typename enable_if<is_unsigned<T>::value && (sizeof(T) <= 4)>::type>
Tuint popcount_u(T i)
{
    return __builtin_popcount(i);
}
inline Tuint popcount_u(Uint64_t i)
{
    return popcount_u(Uint32_t(i)) + popcount_u(Uint32_t(i >> 32));
}

template<typename T, typename = typename enable_if<is_integral<T>::value>::type>
Tuint popcount(T i)
{
    /// \ingroup functions
    /// Counts the number of bits in the given value of integral type
    return popcount_u(static_cast<typename make_unsigned<T>::type>(i));
}

//@{

template<typename T, typename = typename enable_if<is_integral<T>::value && (sizeof(typename unrangetype_core<T>::type) <= 4)>::type>
Tuint clz(T i)
{
    /// \ingroup functions
    /// Counts the number of leading zeros in the given value of integral type
    if (i == 0) return 8 * sizeof(typename unrangetype_core<T>::type);
    if (i < 0)
        return __builtin_clz((Uint32_t)i);
    else
        return __builtin_clz((Uint32_t)i) - 8 * (4 - sizeof(typename unrangetype_core<T>::type));
}
inline Tuint clz(Uint64_t i)
{
    /// \ingroup functions
    /// Counts the number of leading zeros in the given value of integral type
    if ((i >> 32) == 0)
        return 32 + clz(Tuint32_t(i));
    else
        return clz(Tuint32_t(i >> 32));
}
//@}

template<typename I, typename = typename enable_if<is_integral<I>::value>::type>
I rotl(I a, Tint b)
{
    /// \ingroup functions
    /// rotate left
    typename make_unsigned<I>::type au = a;
    Tuint                           bu = b;
    bu                                 = bu % (sizeof(typename unrangetype_core<I>::type) * 8);
    return I((au << bu) | (bu == 0 ? 0 : (au >> (sizeof(typename unrangetype_core<I>::type) * 8 - bu))));
}

template<typename I, typename = typename enable_if<is_integral<I>::value>::type>
I rotr(I a, Tint b)
{
    /// \ingroup functions
    /// rotate right
    typename make_unsigned<I>::type au = a;
    Tuint                           bu = b;
    bu %= (sizeof(typename unrangetype_core<I>::type) * 8);
    return I((au >> bu) | (bu == 0 ? 0 : (au << (sizeof(typename unrangetype_core<I>::type) * 8 - bu))));
}

inline Double sinpi(Double x)
{
    /// \ingroup functions
    /// sin(x*PI)
    x -= impl::round(x / 2) * 2;
    return impl::sin(x * PI);
}
inline Float sinpi(Float x)
{
    /// \ingroup functions
    /// sin(x*PI)
    x -= impl::round(x / 2) * 2;
    return impl::sin(x * PI);
}
inline Half sinpi(Half x)
{
    /// \ingroup functions
    /// sin(x*PI)
    x -= impl::round(x / 2) * 2;
    return impl::sin(x * PI);
}
inline Double cospi(Double x)
{
    /// \ingroup functions
    /// cos(x*PI)
    x -= impl::round(x / 2) * 2;
    return impl::cos(x * PI);
}
inline Float cospi(Float x)
{
    /// \ingroup functions
    /// cos(x*PI)
    x -= impl::round(x / 2) * 2;
    return impl::cos(x * PI);
}
inline Half cospi(Half x)
{
    /// \ingroup functions
    /// cos(x*PI)
    x -= impl::round(x / 2) * 2;
    return impl::cos(x * PI);
}
inline Double tanpi(Double x)
{
    /// \ingroup functions
    /// tan(x*PI)
    x -= impl::round(x);
    if (x == 0.5)
        return numeric_limits<Tfloat>::infinity();
    if (x == -0.5)
        return -numeric_limits<Tfloat>::infinity();
    return impl::tan(x * PI);
}
inline Float tanpi(Float x)
{
    /// \ingroup functions
    /// tan(x*PI)
    x -= impl::round(x);
    if (x == 0.5)
        return numeric_limits<Tfloat>::infinity();
    if (x == -0.5)
        return -numeric_limits<Tfloat>::infinity();
    return impl::tan(x * PI);
}
inline Half tanpi(Half x)
{
    /// \ingroup functions
    /// tan(x*PI)
    x -= impl::round(x);
    if (x == 0.5)
        return numeric_limits<Tfloat>::infinity();
    if (x == -0.5)
        return -numeric_limits<Tfloat>::infinity();
    return impl::tan(x * PI);
}
inline Double asinpi(Double x)
{
    /// \ingroup functions
    /// asin(x) / PI
    return asin(x) / PI;
}
inline Float asinpi(Float x)
{
    /// \ingroup functions
    /// asin(x) / PI
    return asin(x) / PI;
}
inline Half asinpi(Half x)
{
    /// \ingroup functions
    /// asin(x) / PI
    return asin(x) / PI;
}
inline Double acospi(Double x)
{
    /// \ingroup functions
    /// acos(x) / PI
    return acos(x) / PI;
}
inline Float acospi(Float x)
{
    /// \ingroup functions
    /// acos(x) / PI
    return acos(x) / PI;
}
inline Half acospi(Half x)
{
    /// \ingroup functions
    /// acos(x) / PI
    return acos(x) / PI;
}
inline Double atanpi(Double x)
{
    /// \ingroup functions
    /// acos(x) / PI
    return atan(x) / PI;
}
inline Float atanpi(Float x)
{
    /// \ingroup functions
    /// acos(x) / PI
    return atan(x) / PI;
}
inline Half atanpi(Half x)
{
    /// \ingroup functions
    /// acos(x) / PI
    return atan(x) / PI;
}
}

using impl::acos;
using impl::acosh;
using impl::acospi;
using impl::asin;
using impl::asinh;
using impl::asinpi;
using impl::atan;
using impl::atanh;
using impl::atanpi;
using impl::clz;
using impl::cos;
using impl::cosh;
using impl::cospi;
using impl::erf;
using impl::erfc;
using impl::exp;
using impl::exp2;
using impl::log;
using impl::log2;
using impl::popcount;
using impl::rotl;
using impl::rotr;
using impl::sin;
using impl::sinh;
using impl::sinpi;
using impl::tan;
using impl::tanh;
using impl::tanpi;
using impl::tgamma;

}

/// @}
