// -*- C++ -*-

#pragma once


#include "goopax_license.h"
#include "gpu_types.h"


/// \addtogroup threads Thread numbers
/// These functions can be used to query the thread layout, the number of threads, and the id of the current thread. All functions may only be called within a kernel.
///
/// The threads are organized hierarchically:
/// - Programming normally occurs at the level of individual threads.
/// - A warp contains individual threads that execute SIMD instructions. Threads within a warp cannot branch into if clauses or loops without all other threads in the warp waiting. The functions ballot() and shuffle() can be used for fast inner-warp communication
/// - A workgroup contains one or more warps. Threads within a workgroup can communicate via local memory
/// - The total number of threads consists of multiple workgroups. Threads of different workgroups can only communicate via global memory by using atomic access functions.

/// \addtogroup device
/// \{

struct _cl_command_queue;

namespace goopax {
/// Specifies which backend to use. When multiple backends are set, CUDA and METAL will be preferred over CL for devices that support them.
/// \ingroup device
enum envmode
{
    env_unset = 0,
    env_CPU   = 1, ///< CPU backend
    env_CL    = 2, ///< OpenCL backend
    env_CUDA  = 4, ///< CUDA backend
    env_METAL = 8, ///< Metal backend

    env_GPU = 0xfffe,    ///< Any GPU backend
    env_ALL = 0xffffffff ///< Any backend, CPU and GPU
};
inline envmode operator|(envmode a, envmode b)
{
    /// \ingroup device
    return (envmode)((int)a | (int)b);
}


namespace impl {
class goopax_device;
class goopax_device_impl;

namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_device_set(goopax_device_impl*, Uint, Uint) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) Int goopax_device_get(const goopax_device_impl*, Uint) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) const char* goopax_device_get_string(const goopax_device_impl*, Uint) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) Double goopax_device_get_double(const goopax_device_impl*, Uint) noexcept;


GOOPAX_DLL_EXPORT(3.0.0) bool goopax_type_info(const goopax_device_impl*, Uint, Uint) noexcept;

GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_rd_wait_all2(goopax_device_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_rd_finished_all2(bool*, goopax_device_impl* device) noexcept;

GOOPAX_DLL_EXPORT(3.0.0) void goopax_retain_device(goopax_device_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) void goopax_release_device(goopax_device_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) goopax_device_impl* goopax_build_get_current_device() noexcept;
}
}

/// \ingroup device
/// \brief Goopax device.
/// This class is used to reference the physical devices in the system.
class goopax_device
{
    friend inline goopax_device get_device_from_gl_common(envmode backend, const struct glinitdata glinit);
    friend inline goopax_device get_device_from_cl_queue(_cl_command_queue* queue);
    friend goopax_device_impl;


    goopax_device_impl* impl;

    inline void incref() const
    {
        if (impl != nullptr) interface::goopax_retain_device(impl);
    }
    inline void decref() const
    {
        if (impl != nullptr) interface::goopax_release_device(impl);
    }


    public:
    goopax_device_impl* get_impl() const
    {
        return impl;
    }

    /// \brief checks if the device is valid.
    /// This will return false if the device is default-constructed.
    /// or if default_device() did not find
    bool valid() const
    {
        return (get_impl() != nullptr);
    }

    /// returns the backend type that is used to access this device
    inline envmode get_envmode() const { return (envmode)interface::goopax_device_get(impl, 10); }

    /// \name overriding default thread numbers
    ///\{

    /// forces the local size for this device to the given value
    void force_local_size(Tuint value) { check_errno(interface::goopax_E_device_set(impl, 0, value)); } // FIXME: Change default to 0, then make manual
    /// forces the global size for this device to the given value
    void force_global_size(Tuint value) { check_errno(interface::goopax_E_device_set(impl, 1, value)); }
    /// forces the number of workgroups for this device to the given value
    void force_num_groups(Tuint value) { check_errno(interface::goopax_E_device_set(impl, 2, value)); }
    void force_compfac(Tuint value) { check_errno(interface::goopax_E_device_set(impl, 3, value)); } // FIXME: Remove?
    ///\}

    Tuint get_local_size() const { return interface::goopax_device_get(impl, 0); }
    Tuint get_global_size() const { return interface::goopax_device_get(impl, 1); }
    Tuint get_num_groups() const { return interface::goopax_device_get(impl, 2); }
    Tuint get_compfac() const { return interface::goopax_device_get(impl, 3); }

    /// \name default thread numbers
    /// These default values are used if they are not overridden
    /// \{

    /// local size
    inline Tuint default_local_size() const { return interface::goopax_device_get(impl, 4); }
    /// instruction level parallelism
    inline Tuint default_num_subthreads() const { return interface::goopax_device_get(impl, 5); }
    /// number of thread groups
    inline Tuint default_num_groups() const { return interface::goopax_device_get(impl, 6); }
    /// total number of threads. Equals local_size*num_groups
    inline Tuint default_global_size() const { return interface::goopax_device_get(impl, 7); }
    /// Size of local memory
    inline Tuint default_local_mem() const { return interface::goopax_device_get(impl, 8); }
    /// \}

    inline Tuint max_local_size() const { return interface::goopax_device_get(impl, 256); }
    inline Tuint max_num_subthreads() const { return interface::goopax_device_get(impl, 257); }
    inline Tuint max_num_groups() const { return interface::goopax_device_get(impl, 258); }
    inline Tuint max_global_size() const { return interface::goopax_device_get(impl, 259); }

    /// \name floating point support
    ///\{

    /// Returns true if the device supports half precision
    inline Tbool support_type(Half) const { return interface::goopax_type_info(impl, 0, type_enum<Thalf>::value); }
    /// Returns true if the device supports single precision floating points
    inline Tbool support_type(Float) const { return interface::goopax_type_info(impl, 0, type_enum<Tfloat>::value); }
    /// Returns true if the device supports double precision floating points
    inline Tbool support_type(Double) const { return interface::goopax_type_info(impl, 0, type_enum<Tdouble>::value); }
    ///\}
    template<typename T, typename TA>
    inline Tbool support_type(checknan<T, TA>) const
    {
        return support_type(T());
    }

    /// \name denormalization support
    ///\{

    /// Returns true if the device supports denormalized numbers for gpu_half
    inline Tbool support_fp_denorm(Half) const { return interface::goopax_type_info(impl, 1, type_enum<Thalf>::value); }
    /// Returns true if the device supports denormalized numbers for gpu_float
    inline Tbool support_fp_denorm(Float) const { return interface::goopax_type_info(impl, 1, type_enum<Tfloat>::value); }
    /// Returns true if the device supports denormalized numbers for gpu_double
    inline Tbool support_fp_denorm(Double) const { return interface::goopax_type_info(impl, 1, type_enum<Tdouble>::value); }
    ///\}
    template<typename T, typename TA>
    inline Tbool support_fp_denorm(checknan<T, TA>) const
    {
        return support_fp_denorm(T());
    }

    /// Returns true if the device supports shared memory addressing between the host and potentially other devices.
    inline Tbool support_svm() const { return interface::goopax_device_get(impl, 12); }

    /// \name device information
    ///\{

    /// Vendor name
    const char* vendor() const { return interface::goopax_device_get_string(impl, 0); }
    /// device name
    const char* name() const { return interface::goopax_device_get_string(impl, 1); }
    /// max device frequency
    double max_frequency() const { return interface::goopax_device_get_double(impl, 0); }
    /// Returns an estimate for the floating point operations the device can perform per second
    double flops() const { return interface::goopax_device_get_double(impl, 1); }
    /// Returns the warp size of the device. This is typically the SIMD instruction width. Threads within a warp cannot branch into loops and if-clauses without all other threads waiting for them to finish.
    Tuint warp_size() const { return interface::goopax_device_get(impl, 11); }
    ///\}

    /// Waits for all kernels to finish on the device.
    inline void wait_all() const
    {
        check_errno(interface::goopax_E_rd_wait_all2(impl));
    }
    /// Tests if all kernels on the device have finished.
    inline Tbool finished_all() const
    {
        return (if_call(interface::goopax_E_rd_finished_all2, impl));
    }

    /// tests if two devices are the same
    bool operator==(goopax_device b) const { return impl == b.impl; }
    /// tests if two devices are different
    bool operator!=(goopax_device b) const { return impl != b.impl; }

    goopax_device(goopax_device_impl* impl0 = nullptr)
        : impl(impl0)
    {
        incref();
    }
    goopax_device& operator=(const goopax_device& b)
    {
        /// copy operator
        decref();
        impl = b.impl;
        incref();
        return *this;
    }
    goopax_device(const goopax_device& b)
        : impl(b.impl)
    {
        /// copy constructor
        incref();
    }
    goopax_device(goopax_device&& b)
        : impl(b.impl)
    {
        b.impl = nullptr;
    }
    goopax_device& operator=(goopax_device&& b)
    {
        swap(impl, b.impl);
        return *this;
    }
    ~goopax_device()
    {
        decref();
    }
};

inline goopax_device get_current_build_device()
{
    /// \ingroup device
    /// Returns the device for which the current kernel is being built. Only valid while building kernel.
    return interface::goopax_build_get_current_device();
}


namespace gpu_thread {
inline gpu_uint local_id();
inline Tuint    warp_size();
inline Tuint    local_size();


/// \name querying the current thread id
/// \{

inline gpu_uint thread_id_in_warp()
{
    /// \ingroup threads
    /// Returns the id in the current warp. May only be called in a GPU kernel.
    return local_id() & (warp_size() - 1);
}
inline gpu_uint warp_id_in_group()
{
    /// \ingroup threads
    /// Returns the warp id in the current group. May only be called in a GPU kernel.
    return local_id() / warp_size();
}
GOOPAX_DEPRECATED inline gpu_uint warp_id() // Deprecated due to ambiguity. Use thread_id_in_warp() or warp_id_in_group() instead.
{
    /// \ingroup threads
    /// Returns the id in the current warp. May only be called in a GPU kernel.
    return thread_id_in_warp();
}
inline gpu_uint local_id()
{
    /// \ingroup threads
    /// Returns the local thread ID in the workgroup. May only be called in a GPU kernel.
    gpu_uint ret;
    interface::goopax_thread_local_id2(ret.impl.get());
    return ret;
}
inline gpu_uint group_id()
{
    /// \ingroup threads
    /// Returns the workgroup ID. May only be called in a GPU kernel.
    gpu_uint ret;
    interface::goopax_thread_group_id2(ret.impl.get());
    return ret;
}
inline gpu_uint global_id()
{
    /// \ingroup threads
    /// Returns the total thread ID on the GPU. Equals group_id()*local_size() + local_id(). May only be called in a GPU kernel.
    return group_id() * local_size() + local_id();
}
/// \}

/// \name querying the thread layout of the current kernel
/// \{

inline Tuint warp_size()
{
    /// \ingroup threads
    /// Returns the warp size of the device for which the kernel is being built. May only be called in a GPU kernel.
    /// The warp size is typically the SIMD instruction width. Threads within a warp cannot branch into loops and if-clauses without all other threads waiting for them to finish.
    return get_current_build_device().warp_size();
}
inline Tuint local_size()
{
    /// \ingroup threads
    /// Returns the workgroup size. May only be called in a GPU kernel
    return if_call(interface::goopax_E_thread_local_size2);
}
inline Tuint num_groups()
{
    /// \ingroup threads
    /// Returns the number of workgroups. May only be called in a GPU kernel
    return if_call(interface::goopax_E_thread_num_groups2);
}
inline Tuint global_size()
{
    /// \ingroup threads
    /// Returns the total number of threads on the GPU. Equals local_size()*num_groups(). May only be called in a GPU kernel
    return if_call(interface::goopax_E_thread_global_size2);
}

inline Tuint num_subthreads()
{
    /// \ingroup threads
    /// Returns the instruction level parallelization that works best for the GPU. May only be called in a GPU kernel
    return if_call(interface::goopax_E_thread_num_subthreads2);
}
inline Tsize_t max_local_mem()
{
    /// \ingroup threads
    /// Returns the maximum amount of local memory for each workgroup. May only be called in a GPU kernel
    return if_call(interface::goopax_E_thread_max_local_mem2);
}
/// \}
}


}
using impl::get_current_build_device;
using namespace impl::gpu_thread;
}

/// \}
