// -*- C++ -*-

#pragma once

#include "gpu_types.h"


namespace goopax {
namespace impl {
template<class T>
struct gather_base;

/// \addtogroup gather Gather return values
/// gather types. These types can be used to gather values from the GPU threads to return values from GPU kernels.
/// Some types are predefined here. User-defined types can be added accordingy. See the predefined types in gather.h for how this is done.
/// If used as return type or reference argument of the program() function of GPU kernels, values are returned as goopax_future when the kernel is called.
///\{

/// \ingroup gather
/// returns the sum of all values.
template<class T>
struct gather_add : public gather_base<T>
{
    using gpu_T = typename gettype<T>::gpu;
    static T init()
    {
        /// null element of the operation, used by buffer::sum() operation.
        return 0;
    }
    template<class U>
    static inline U op(const U& a, const U& b)
    {
        /// here the operation is defined.
        return a + b;
    }
    gpu_T& operator=(const gpu_T& v)
    {
        return (static_cast<gpu_T&>(*this) = v);
    }
    gather_add()
    {}
    gather_add(const gpu_T& val)
        : gather_base<T>(val)
    {}
};

/// \ingroup gather
/// returns the minimum of all values
template<class T>
struct gather_min : public gather_base<T>
{
    using gpu_T = typename gettype<T>::gpu;
    static T init()
    {
        return numeric_limits<T>::max();
    }
    template<class U>
    static inline U op(const U& a, const U& b)
    {
        return min(a, b);
    }
    gpu_T& operator=(const gpu_T& v)
    {
        return (static_cast<gpu_T&>(*this) = v);
    }
    gather_min()
    {}
    gather_min(const gpu_T& val)
        : gather_base<T>(val)
    {}
};

/// \ingroup gather
/// returns the maximum of all values
template<class T>
struct gather_max : public gather_base<T>
{
    using gpu_T = typename gettype<T>::gpu;
    static GOOPAX_CONSTEXPR T init()
    {
        return (numeric_limits<T>::is_integer ? numeric_limits<T>::min() : -numeric_limits<T>::max());
    }
    template<class U>
    static inline U op(const U& a, const U& b)
    {
        return max(a, b);
    }
    gpu_T& operator=(const gpu_T& v)
    {
        return (static_cast<gpu_T&>(*this) = v);
    }
    gather_max() {}
    gather_max(const gpu_T& val)
        : gather_base<T>(val)
    {}
};
/// \}

}
using impl::gather_add;
using impl::gather_max;
using impl::gather_min;
}
