// -*- C++ -*-

#pragma once

#include "buffer.h"
#include "gpu-utils-core.h"


namespace goopax {
namespace impl {
template<class OP, typename size_type>
struct minmaxsum_prog :
#if GOOPAX_USE_CXX14
    public kernel<minmaxsum_prog<OP, size_type>>
#else
    public kernel<minmaxsum_prog<OP, size_type>, typename OP::value_type>
#endif
{
    using gpu_size_type = typename gettype<size_type>::gpu;

    static minmaxsum_prog& static_prog()
    {
        static minmaxsum_prog prog;
        return prog;
    }

    using T = typename OP::value_type;
    typedef typename gettype<T>::gpu gpu_T;

    OP program(const resource<T, size_type>& src,
               const gpu_size_type           start,
               const gpu_size_type           end)
    {
        const Uint32_t blocksize = 4;

        gpu_T res = OP::init();

        if (blocksize == 1)
        {
            gpu_for_global(gpu_size_type src_p = start, end)
            {
                res = OP::op(res, (gpu_T)src[src_p]);
            }
        }
        else
        {
            const gpu_size_type startblock = (start + (blocksize - 1)) & (-(Tint32_t)blocksize);
            const gpu_size_type endblock   = max(startblock, end & (-(Tint32_t)blocksize));

            gpu_for_global(gpu_size_type src_p = start, min(startblock, end))
            {
                res = OP::op(res, (gpu_T)src[src_p]);
            }

            gpu_for(gpu_size_type src_p = startblock + (global_id() * blocksize), endblock, global_size() * blocksize)
            {
                for (Tuint sub = 0; sub < blocksize; ++sub)
                {
                    res = OP::op(res, (gpu_T)src[src_p + sub]);
                }
            }

            gpu_for_global(gpu_size_type src_p = endblock, end)
            {
                res = OP::op(res, (gpu_T)src[src_p]);
            }
        }

        return res;
    }

    virtual ~minmaxsum_prog() {}
};


template<class DEST, class SRC, Int burstsize>
struct gather_prog : public kernel<gather_prog<DEST, SRC, burstsize>>
{
    using size_type     = typename SRC::size_type;
    using gpu_size_type = typename gettype<size_type>::gpu;

    static gather_prog& static_prog()
    {
        static gather_prog prog;
        return prog;
    }

    void program(typename DEST::resource_type&      dest,
                 const typename SRC::resource_type& src,
                 const gpu_size_type                NumThreads,
                 const gpu_size_type                NumBlocks,
                 const gpu_size_type                start_dest,
                 const gpu_size_type                start_src)
    {
        typedef typename DEST::value_type T;
        typedef typename gettype<T>::gpu  gpu_T;

        {
            gpu_for_global(gpu_size_type block = 0,
                           NumBlocks)
            {
                array<gpu_T, burstsize> sum;
                sum.fill(gpu_T(0));
                gpu_for(gpu_uint thread = 0, NumThreads)
                {
                    for (Tsize_t s = 0; s < burstsize; ++s)
                    {
                        sum[s] += src[start_src + NumThreads * block * burstsize + thread * burstsize + s];
                    }
                }
                for (Tsize_t s = 0; s < burstsize; ++s)
                {
                    dest[start_dest + block * burstsize + s] = sum[s];
                }
            }
        }
    }

    virtual ~gather_prog()
    {}
};


namespace utils {
template<typename T, typename size_type>
T min(const buffer<T, size_type>& ar, const size_type begin, const size_type end)
{
    return minmaxsum_prog<gather_min<T>, size_type>::static_prog()(ar, begin, end).get();
}
template<typename T, typename size_type>
T max(const buffer<T, size_type>& ar, const size_type begin, const size_type end)
{
    return minmaxsum_prog<gather_max<T>, size_type>::static_prog()(ar, begin, end).get();
}
template<typename T, typename size_type>
T sum(const buffer<T, size_type>& ar, const size_type begin, const size_type end)
{
    return minmaxsum_prog<gather_add<T>, size_type>::static_prog()(ar, begin, end).get();
}

}


namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_assert_impl2(gpu_writer_base_impl* cond, const char* FILE, Size_t LINE, const char* info) noexcept;
}
}

inline void gpu_assert_impl(gpu_bool cond, const char* FILE, Tsize_t LINE, const char* info)
{
    check_errno(interface::goopax_E_assert_impl2(cond.impl.get(), FILE, LINE, info));
}

/// \ingroup debug
/// asserts that the given value of type gpu_bool is true.
/// If not, an error message is displayed, and an exception is thrown.
#define gpu_assert(...) ::goopax::impl::gpu_assert_impl((__VA_ARGS__), __FILE__, __LINE__, #__VA_ARGS__);


}
}
