// -*- C++ -*-

#pragma once

#include "env.h"
#include "memory.h"


struct _cl_mem;

/// \addtogroup image
/// \{

namespace goopax {
/// image sampler modes.
enum sampler_mode
{
    coords_normalized = 1,

    address_mirrored_repeat = 2,
    address_repeat          = 4,
    address_clamp_to_edge   = 8,
    address_clamp           = 16,
    address_none            = 32,

    filter_nearest = 64,
    filter_linear  = 128
};

inline sampler_mode operator|(sampler_mode a, sampler_mode b)
{
    return (sampler_mode)((int)a | (int)b);
}

namespace impl {
namespace ns_param_kernel {
template<typename BUFREF, class R, typename IMAGE, class REST>
struct pas_common_image;
}


template<typename T, Size_t N>
array<T, N> makearray(T fill)
{
    array<T, N> ret;
    ret.fill(fill);
    return ret;
}

template<typename T, Size_t NA, Size_t NB>
array<T, NA + NB> concatenate(const array<T, NA>& a, const array<T, NB>& b)
{
    array<T, NA + NB> ret;
    std::copy(a.begin(), a.end(), ret.begin());
    std::copy(b.begin(), b.end(), ret.begin() + a.size());
    return ret;
}


namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(3.0.0) const Tuint* goopax_image_dimensions2(const texbuf*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_image_create2(texbuf**, goopax_device_impl*, void (*)(texbuf&, texbuf&), Uint, const Tuint*, Bool, const Tuint*, IMAGE_FLAGS, void*, Uint, Uint, Bool) noexcept;
#if GOOPAX_HAVE_OPENGL
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_image_create_gl2(texbuf**, goopax_device_impl* device, void (*)(texbuf&, texbuf&), Uint, Bool, Uint, Uint, Int, IMAGE_FLAGS, Uint, Uint, Bool) noexcept;
#endif
#if GOOPAX_HAVE_OPENCL
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_image_create_cl2(texbuf**, goopax_device_impl* device, void (*)(texbuf&, texbuf&), Uint, Bool, _cl_mem*, Uint, Uint, Bool) noexcept;
#endif

GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_image_resource_create2(resource_impl**, Uint, Bool, texbuf*, Bool, Uint, Uint, Bool) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_image_write2(resource_impl*, const gpu_writer_base_impl* const*, const gpu_writer_base_impl* const*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_image_read3(resource_impl*, const gpu_writer_base_impl* const*, gpu_writer_base_impl* const*, sampler_mode) noexcept;
GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_image_lock_map(texbuf*, void**, Tuint*) noexcept;
GOOPAX_DLL_EXPORT(3.1.1) errno_type goopax_E_image_lock_map2(texbuf*, const Tuint* origin, const Tuint* region, void**, Tuint*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_copy_image2(texbuf*, texbuf*, const Tuint*, const Tuint*, const Tuint*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_copy_image_to_host2(texbuf*, void*, const Tuint*, const Tuint*, const Tuint*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_copy_image_from_host2(texbuf*, const void*, const Tuint*, const Tuint*, const Tuint*) noexcept;
GOOPAX_DLL_EXPORT(3.0.1) errno_type goopax_E_fill_image(texbuf*, const void*, const Tuint*, const Tuint*) noexcept;
}
}


template<typename T, Bool is_signed, Bool is_integral, Bool normalized>
struct get_channel_access_type;

template<typename T, Bool is_signed, Bool is_integral>
struct get_channel_access_type<T, is_signed, is_integral, true>
{
    using type = gpu_float;
};
template<typename T>
struct get_channel_access_type<T, true, true, false>
{
    using type = gpu_int;
};
template<typename T>
struct get_channel_access_type<T, false, true, false>
{
    using type = gpu_uint;
};
template<typename T>
struct get_channel_access_type<T, true, false, false>
{
    static_assert(is_floating_point<T>::value, "");
    using type = gpu_float;
};


template<typename T>
struct channel_type_info
{
    using TU                                = typename unrangetype<T>::type;
    using channel_type                      = T;
    static constexpr Uint channel_type_enum = type_enum<T>::value;
    template<Bool normalized>
    using channel_access_type                = typename get_channel_access_type<TU, is_signed<T>::value, is_integral<T>::value, normalized>::type;
    static constexpr Bool allow_normalized   = is_integral<T>::value && (gettype<T>::size <= 2);
    static constexpr Bool allow_unnormalized = is_integral<T>::value || is_floating_point<T>::value || is_half<T>::value;
    static constexpr Bool allow_sampler(Bool normalized) { return normalized || is_floating_point<T>::value || is_half<T>::value; }
};

template<typename T>
struct image_type_info : public channel_type_info<typename gpu_struct::type_probe<T>::type>
{
    static constexpr Uint num_channels = gpu_struct::type_probe<T>::N;
    static_assert(num_channels == 4, "only 4 channels are supported");
};


template<Uint DIM, typename T, Bool normalized = !image_type_info<T>::allow_unnormalized>
class image_buffer;
template<Uint DIM, typename T, Bool normalized = !image_type_info<T>::allow_unnormalized>
class image_array_buffer;
template<Uint DIM, typename T, Bool normalized = !image_type_info<T>::allow_unnormalized>
class image_resource;
template<Uint DIM, typename T, Bool normalized = !image_type_info<T>::allow_unnormalized>
class image_array_resource;
template<Uint DIM, typename T, Bool CONST>
class image_array_buffer_map;

template<typename P, typename TO>
struct change_ptr;
template<typename T, typename TO>
struct change_ptr<T*, TO>
{
    using type = TO*;
};
template<typename T, typename TO>
struct change_ptr<const T*, TO>
{
    using type = const TO*;
};


/// \ingroup image
/// Maps an image buffer to host memory and provides access to the image from the host.
template<Uint DIM, typename T, Bool CONST = false>
class image_buffer_map : private map_base
{
    friend class image_array_buffer_map<DIM, T, CONST>;

    public:
    using value_type                     = T;
    static constexpr Uint num_dimensions = DIM;
    static constexpr Bool is_array       = false;

    private:
    T*                    ptr = nullptr;
    array<Tuint, DIM - 1> elepitchdim;
    array<Tuint, DIM>     dimensions_i;

    public:
    array<Tuint, DIM> dimensions() const
    {
        /// dimensions of image map
        return dimensions_i;
    }

    template<typename EIF = void, typename = typename enable_if<!CONST, EIF>::type>
    value_type& operator[](array<Tuint, DIM> pos)
    {
        /// image access from host
        /// \param pos position in image
        /// \return reference to image pixel
        Tsize_t linpos = pos[DIM - 1];

        for (Tint k = DIM - 2; k >= 0; --k)
            linpos = linpos * elepitchdim[k] + pos[k];

        return ptr[linpos];
    }
    const value_type& operator[](array<Tuint, DIM> pos) const
    {
        /// image access from host
        /// \param pos position in image
        /// \return reference to image pixel
        Tsize_t linpos = pos[DIM - 1];

        for (Tint k = DIM - 2; k >= 0; --k)
            linpos = linpos * elepitchdim[k] + pos[k];

        return ptr[linpos];
    }

    template<Bool normalized, typename EIF = void, typename = typename enable_if<!CONST, EIF>::type>
    image_buffer_map(image_buffer<DIM, T, normalized>& buf)
        : image_buffer_map(buf, makearray<Tuint, DIM>(0), buf.dimensions())
    {
        /// constructor.
        /// \param buf image that is to be mapped to host memory
    }

    template<Bool normalized, typename EIF = void, typename = typename enable_if<!CONST, EIF>::type>
    image_buffer_map(image_buffer<DIM, T, normalized>& buf, array<Tuint, DIM> origin, array<Tuint, DIM> region)
        : dimensions_i(buf.dimensions())
    {
        /// constructor.
        /// \param buf image that is to be mapped to host memory
        /// \param origin begin of image region
        /// \param region size of image region
        void* tmp;
        check_errno(interface::goopax_E_image_lock_map2(buf.get_buf(), origin.data(), region.data(), &tmp, elepitchdim.data()));
        ptr       = static_cast<T*>(tmp);
        this->buf = buf.get_buf();
    }

    template<Bool normalized, typename EIF = void, typename = typename enable_if<CONST, EIF>::type>
    image_buffer_map(const image_buffer<DIM, T, normalized>& buf)
        : image_buffer_map(buf, makearray<Tuint, DIM>(0), buf.dimensions())
    {
        /// constructor.
        /// \param buf const image that is to be mapped to host memory.
    }
    template<Bool normalized, typename EIF = void, typename = typename enable_if<CONST, EIF>::type>
    image_buffer_map(const image_buffer<DIM, T, normalized>& buf, array<Tuint, DIM> origin, array<Tuint, DIM> region)
        : dimensions_i(buf.dimensions())
    {
        /// constructor.
        /// \param buf image that is to be mapped to host memory
        /// \param origin begin of image region
        /// \param region size of image region
        void* tmp;
        check_errno(interface::goopax_E_image_lock_map2(buf.get_buf(), origin.data(), region.data(), &tmp, elepitchdim.data()));
        ptr       = static_cast<T*>(tmp);
        this->buf = buf.get_buf();
    }


    private:
    image_buffer_map() = default;
};
#if GOOPAX_CONSTEXPR_INSTANTIATION
template<Uint DIM, typename T, Bool CONST>
const Uint image_buffer_map<DIM, T, CONST>::num_dimensions;
template<Uint DIM, typename T, Bool CONST>
const Bool image_buffer_map<DIM, T, CONST>::is_array;
#endif

/// \ingroup image const image buffer map that can be used to map const references of images.
template<Uint DIM, typename T>
using const_image_buffer_map = image_buffer_map<DIM, T, true>;


/// \ingroup image
/// Maps an image buffer array to host memory
template<Uint DIM, typename T, Bool CONST = false>
class image_array_buffer_map : private map_base
{
    public:
    using value_type                     = T;
    static constexpr Uint num_dimensions = DIM;
    static constexpr Bool is_array       = true;

    private:
    T*                    ptr = nullptr;
    array<Tuint, DIM>     elepitchdim;
    array<Tuint, DIM + 1> dimensions_i;
    //Tsize_t array_size_i;

    public:
    array<Tuint, DIM> dimensions() const
    {
        /// image dimensions of the image map.
        array<Tuint, DIM> ret;
        std::copy(dimensions_i.begin(), dimensions_i.begin() + DIM, ret.begin());
        return ret;
    }
    Tsize_t array_size() const
    {
        /// number of images in the array map
        return dimensions_i.back();
    }

    image_buffer_map<DIM, T, CONST> operator[](Tsize_t slice)
    {
        /// accessing image slice in the image array
        /// \param slice image number in the array
        /// \return image map to the image
        Tsize_t slicepitch = 1;
        for (Tuint k = 0; k < DIM; ++k)
        {
            slicepitch *= elepitchdim[k];
        }
        image_buffer_map<DIM, T, CONST> ret;
        ret.buf = nullptr;
        ret.ptr = this->ptr + slicepitch * slice;
        std::copy(elepitchdim.begin(), elepitchdim.begin() + DIM - 1, ret.elepitchdim.begin());
        ret.dimensions_i = dimensions();
        return ret;
    }
    image_buffer_map<DIM, T, true> operator[](Tsize_t slice) const
    {
        /// accessing image slice in the image array
        /// \param slice image number in the array
        /// \return image map to the image
        Tsize_t slicepitch = 1;
        for (Tuint k = 0; k < DIM; ++k)
        {
            slicepitch *= elepitchdim[k];
        }
        image_buffer_map<DIM, T, true> ret;
        ret.buf = nullptr;
        ret.ptr = this->ptr + slicepitch * slice;
        std::copy(elepitchdim.begin(), elepitchdim.begin() + DIM - 1, ret.elepitchdim.begin());
        ret.dimensions_i = dimensions();
        return ret;
    }


    template<Bool normalized, typename EIF = void, typename = typename enable_if<!CONST, EIF>::type>
    image_array_buffer_map(image_array_buffer<DIM, T, normalized>& buf)
        : image_array_buffer_map(buf, makearray<Tuint, DIM + 1>(0), concatenate(buf.dimensions(), array<Tuint, 1>({ { buf.array_size() } })))
    {
        /// constructor.
        /// \param buf image array that is to be mapped to host memory
    }

    template<Bool normalized, typename EIF = void, typename = typename enable_if<!CONST, EIF>::type>
    image_array_buffer_map(image_array_buffer<DIM, T, normalized>& buf, array<Tuint, DIM + 1> origin, array<Tuint, DIM + 1> region)
        : dimensions_i(region)
    {
        /// constructor.
        /// \param buf image array that is to be mapped to host memory
        /// \param origin of image region. The last element is the first array index
        /// \param region size of image region. The last element specifies the number of images.
        void* tmp;
        check_errno(interface::goopax_E_image_lock_map2(buf.get_buf(), origin.data(), region.data(), &tmp, elepitchdim.data()));
        ptr       = static_cast<T*>(tmp);
        this->buf = buf.get_buf();
    }

    template<Bool normalized, typename EIF = void, typename = typename enable_if<CONST, EIF>::type>
    image_array_buffer_map(const image_array_buffer<DIM, T, normalized>& buf)
        : image_array_buffer_map(buf, makearray<Tuint, DIM + 1>(0), concatenate(buf.dimensions(), array<Tuint, 1>({ { buf.array_size() } })))
    {
        /// constructor.
        /// \param buf image array that is to be mapped to host memory
    }

    template<Bool normalized, typename EIF = void, typename = typename enable_if<CONST, EIF>::type>
    image_array_buffer_map(const image_array_buffer<DIM, T, normalized>& buf, array<Tuint, DIM + 1> origin, array<Tuint, DIM + 1> region)
        : dimensions_i(region)
    {
        /// constructor.
        /// \param buf image array that is to be mapped to host memory
        /// \param origin of image region. The last element is the first array index
        /// \param region size of image region. The last element specifies the number of images.
        void* tmp;
        check_errno(interface::goopax_E_image_lock_map2(buf.get_buf(), origin.data(), region.data(), &tmp, elepitchdim.data()));
        ptr       = static_cast<T*>(tmp);
        this->buf = buf.get_buf();
    }
};
#if GOOPAX_CONSTEXPR_INSTANTIATION
template<Uint DIM, typename T, Bool CONST>
const Uint image_array_buffer_map<DIM, T, CONST>::num_dimensions;
template<Uint DIM, typename T, Bool CONST>
const Bool image_array_buffer_map<DIM, T, CONST>::is_array;
#endif

template<Uint DIM, typename T>
using const_image_array_buffer_map = image_array_buffer_map<DIM, T, true>;

/// \ingroup image
template<Uint DIM, Bool IS_ARRAY, typename T, Bool normalized>
struct image_buffer_base : private init_caller<Tint>
    , public tex_arg_base
    , public image_type_info<T>
{
    using value_type     = T;
    using cpu_size_type  = Tuint;
    using cpu_ssize_type = typename make_signed<cpu_size_type>::type;
    using gpu_size_type  = typename gettype<cpu_size_type>::gpu;
    using gpu_ssize_type = typename gettype<cpu_ssize_type>::gpu;
    using size_type      = cpu_size_type;
    using ssize_type     = cpu_ssize_type;

    using iterator        = T*;
    using const_iterator  = const T*;
    using reference       = T&;
    using const_reference = const T&;

    static const Uint num_dimensions = DIM;
    static const Bool is_checknan    = !is_same<T, typename goopax::impl::unrangetype<T>::type>::value;
    static const Bool is_array       = IS_ARRAY;
    static_assert(normalized ? image_type_info<T>::allow_normalized : image_type_info<T>::allow_unnormalized, "(un)normalized mode not allowed for this image type");

    public:
    inline array<size_type, DIM> dimensions() const
    {
        /// image dimensions
        const Tuint*          p = interface::goopax_image_dimensions2(this->buf);
        array<size_type, DIM> ret;
        for (Tuint k = 0; k < DIM; ++k)
            ret[k] = p[k];
        return ret;
    }
    inline size_type width() const { return dimensions()[0]; };
    template<typename EIF = void, typename = typename enable_if<(DIM >= 2), EIF>::type>
    inline size_type height() const
    {
        return dimensions()[1];
    }
    template<typename EIF = void, typename = typename enable_if<(DIM >= 3), EIF>::type>
    inline size_type depth() const
    {
        return dimensions()[2];
    }

    inline texbuf* get_buf() const
    {
        return this->buf;
    }

    image_buffer_base()
    {}

    image_buffer_base(goopax_device                          device,
                      const array<size_type, DIM + is_array> size,
                      IMAGE_FLAGS                            flags,
                      value_type*                            host_ptr,
                      array<size_type, DIM + is_array - 1>   pitchdim)
        : tex_arg_base(nullptr)
    {
        if (host_ptr == nullptr)
            pitchdim.fill(size_type());

        this->buf = if_call(interface::goopax_E_image_create2, device.get_impl(), nullptr, DIM, size.data(), (Bool)is_array, pitchdim.data(), flags, host_ptr, this->channel_type_enum, this->num_channels, normalized);
    }

    image_buffer_base(INITTEX_IMPL, texbuf* buf)
        : tex_arg_base(buf)
    {}

    image_buffer_base(image_buffer_base&&) = default;
    image_buffer_base& operator=(image_buffer_base&&) = default;
};


template<Uint DIM, Bool IS_ARRAY, typename T, Bool normalized>
const Uint image_buffer_base<DIM, IS_ARRAY, T, normalized>::num_dimensions;

/**
     /// \ingroup image
        Image container. The image will be allocated on the device.
        \tparam DIM: Number of dimensions. Must be 1, 2, od 3.
        \tparam T: Element data type. Must be std::array<C, N>, or other goopax struct of identical types. C: channel type (int32_t, int16_t, int8_t, uint32_t, uint16_t, uint8_t, float, half), N: number of channels (must be 4 at the moment).
     \tparam normalized if set to true, image values are normalized to floats in the range 0..1 for unsigned channel types, or -1..1 for signed channel types. Defaults to false if the channel data type allows unnormalized access
     */
template<Uint DIM, typename T, Bool normalized>
class image_buffer : private image_buffer_base<DIM, false, T, normalized>
{
    using base_t = image_buffer_base<DIM, false, T, normalized>;
    static_assert(DIM >= 1 && DIM <= 3, "Only 1D, 2D, and 3D images are supported.");

    public:
    using resource_type = image_resource<DIM, T, normalized>;

    using base_t::depth; ///< depth of the image
    using base_t::dimensions;
    using base_t::get_buf;
    using base_t::get_device; ///< returns the device where the image is allocated
    using base_t::get_hash;
    using base_t::get_hash_ignore_uninit;
    using base_t::height; ///< height of the image
    using base_t::is_array;
    using base_t::is_checknan;
    using base_t::num_dimensions; ///< number of dimensions
    using base_t::width;          ///< width of the image
    using typename base_t::size_type;
    using typename base_t::value_type; ///< pixe element type

    using map_type       = image_buffer_map<DIM, T>;
    using const_map_type = const_image_buffer_map<DIM, T>;
    base_t& get_base() { return *this; }

    inline goopax_device get_device() const ///< returns the device that the image is allocated on
    {
        return base_t::get_device();
    }

    /**
         copy image region from another image of the same type.
         \param b: source image
         \param src_origin: origin in source image
         \param dest_origin: origin in *this
         \param region: size of image region to copy
         */
    void copy(const image_buffer&   b,
              array<size_type, DIM> src_origin,
              array<size_type, DIM> dest_origin,
              array<size_type, DIM> region)
    {
        check_errno(interface::goopax_E_copy_image2(get_buf(), b.get_buf(), src_origin.data(), dest_origin.data(), region.data()));
    }

    /**
         copy image from another image with the same type and dimensions.
         */
    void copy(const image_buffer& b)
    {
        if (dimensions() != b.dimensions())
        {
            exwhat() << "Different image dimensions while copying image.";
            throw EX::memory();
        }
        array<size_type, DIM> null;
        null.fill(0);

        copy(b, null, null, dimensions());
    }

    //@{
    /**
         copy image region to host.
         \param p: destination address in host memory
         \param origin: origin of image in device
         \param region: size of image region
         \param pitchdim: row pitch dimensions (2D images) or row pitch+slice pitch dimensions (3D images), in pixels.
	*/
    void copy_to_host(T* p, array<size_type, DIM> origin, array<size_type, DIM> region,
                      array<size_type, DIM - 1> pitchdim = makearray<size_type, DIM - 1>(0)) const
    {
        check_errno(interface::goopax_E_copy_image_to_host2(this->get_buf(), p, origin.data(), region.data(), pitchdim.data()));
    }

    void copy_to_host(T* p) const
    {
        array<size_type, DIM> origin;
        origin.fill(0);
        copy_to_host(p, origin, dimensions());
    }
    //@}

    /**
         copy image region from host.
         \param p: source address in host memory
         \param origin: origin of image in device
         \param region: size of image region
         \param pitchdim: row pitch dimensions (2D images) or row pitch+slice pitch dimensions (3D images), in pixels. If values are 0, the pitch is calculated from region
	*/
    void copy_from_host(const T* p, array<size_type, DIM> origin, array<size_type, DIM> region,
                        array<size_type, DIM - 1> pitchdim = makearray<size_type, DIM - 1>(0))
    {
        check_errno(interface::goopax_E_copy_image_from_host2(this->get_buf(), p, origin.data(), region.data(), pitchdim.data()));
    }

    /** copy whole image from host.
         \param p: source address in host memory
         */
    void copy_from_host(const T* p)
    {
        array<size_type, DIM> origin;
        origin.fill(0);
        copy_from_host(p, origin, dimensions());
    }

    void fill(T value, array<size_type, DIM> origin, array<size_type, DIM> region)
    {
        /// FIll image array
        /// \param value color to fill
        /// \param origin origin in destination image.
        /// \param region size of image patch to fill.
        check_errno(interface::goopax_E_fill_image(this->get_buf(), &value, origin.data(), region.data()));
    }
    void fill(T value)
    {
        /// FIll image array
        /// \param value color to fill
        array<size_type, DIM> origin;
        origin.fill(0);
        fill(value, origin, dimensions());
    }


    image_buffer(goopax_device               device,
                 const array<size_type, DIM> size,
                 IMAGE_FLAGS                 flags    = IMAGE_READ_WRITE,
                 value_type*                 host_ptr = nullptr,
                 array<size_type, DIM - 1>   pitchdim = makearray<size_type, DIM - 1>(0))
        : base_t(device, size, flags, host_ptr, pitchdim)
    {
        /**
           Allocated image on the device.
           \param device: device to allocate image on
           \param size: size of image
           \param flags: must be one of IMAGE_READ_WRITE, IMAGE_READ_ONLY, IMAGE_WRITE_ONLY, IMAGE_WRITE_DISCARD
           \param host_ptr: if used, image will be linked to the specified address in host memory.
           \param pitchdim: pitch size in pixels
           */
    }

    /**
         default constructor. Image must not be used until replaced by a real image.
         */
    image_buffer()
    {}

    private:
    image_buffer(INITTEX_IMPL, texbuf* buf)
        : base_t(INITTEX_IMPL(), buf)
    {}

    public:
    static inline image_buffer create_from_gl(goopax_device device,
                                              Uint          GLres,
                                              IMAGE_FLAGS   flags    = IMAGE_READ_WRITE,
                                              Uint          GLtarget = (DIM == 1 ? 0x0DE0 : (DIM == 2 ? 0x0DE1 : 0x806F)) /* GL_TEXTURE_?D */,
                                              Int           miplevel = 0);

    static inline image_buffer create_from_cl(goopax_device device, _cl_mem* mem);
};


/**
     /// \ingroup image
       Image array. The image will be allocated on the device.
        \tparam DIM: Number of dimensions. Must be 1 or 2.
        \tparam T: Element data type. Must be std::array<C, N>, or other goopax struct of identical types. C: channel type (int32_t, int16_t, int8_t, uint32_t, uint16_t, uint8_t, float, half), N: number of channels (must be 4 at the moment).
     \tparam normalized if set to true, image values are normalized to floats in the range 0..1 for unsigned channel types, or -1..1 for signed channel types. Defaults to false if the channel data type allows unnormalized access
     */
template<Uint DIM, typename T, Bool normalized>
class image_array_buffer : private image_buffer_base<DIM, true, T, normalized>
{
    static_assert(DIM == 1 || DIM == 2, "Only 1D and 2D image arrays are supported.");
    using base_t = image_buffer_base<DIM, true, T, normalized>;

    public:
    using resource_type = image_array_resource<DIM, T, normalized>;

    using typename base_t::size_type;
    using typename base_t::value_type;
    /// Number of dimensions
    using base_t::num_dimensions;
    /// dimensions of image
    using base_t::dimensions;
    /// width
    using base_t::width;
    /// height
    using base_t::height;
    /// depth
    using base_t::depth;
    /// device
    using base_t::get_buf;
    using base_t::get_device;
    using base_t::get_hash;
    using base_t::get_hash_ignore_uninit;
    using base_t::is_array;
    using base_t::is_checknan;

    private:
    static inline array<size_type, DIM + 1> flat_dim(size_type imagenum, array<size_type, DIM> p)
    {
        array<size_type, DIM + 1> p2;
        for (Tuint k = 0; k < DIM; ++k) p2[k] = p[k];
        p2.back() = imagenum;
        return p2;
    }

    public:
    using map_type       = image_array_buffer_map<DIM, T>;
    using const_map_type = const_image_array_buffer_map<DIM, T>;

    size_type array_size() const
    {
        /// Number of images in the array
        return interface::goopax_image_dimensions2(this->buf)[DIM];
    }

    /// Reference to single image in the image array
    template<Bool CONST>
    struct image_slice
    {
        typename conditional<CONST, const image_array_buffer&, image_array_buffer&>::type buf;
        size_type                                                                         imagenum;

        inline size_type width() const
        {
            return buf.width();
        }
        template<typename EIF = void, typename = typename enable_if<(DIM >= 2), EIF>::type>
        inline size_type height() const
        {
            return buf.height();
        }

        template<Bool C2, typename EIF = void, typename = typename enable_if<!CONST, EIF>::type>
        inline void copy(const image_slice<C2>& b,
                         array<size_type, DIM>  src_origin,
                         array<size_type, DIM>  dest_origin,
                         array<size_type, DIM>  region)
        {
            /// copy from other image slice
            /// \param src_origin offset in source image
            /// \param dest_origin offset in destination image
            /// \param region size of image patch to copy
            buf.copy(b.buf, flat_dim(b.imagenum, src_origin), flat_dim(imagenum, dest_origin), flat_dim(1, region));
        }

        template<Bool C2, typename EIF = void, typename = typename enable_if<!CONST, EIF>::type>
        inline void copy(const image_slice<C2>& b)
        {
            /// copy from other image size
            if (buf.dimensions() != b.buf.dimensions())
            {
                exwhat() << "Different image dimensions while copying image slice.";
                throw EX::memory();
            }
            array<size_type, DIM> null;
            null.fill(0);
            copy(b, null, null, buf.dimensions());
        }


        // FIXME: Is this DIM+1 thingy ok?
        void copy_to_host(T* p, array<size_type, DIM> origin, array<size_type, DIM> region,
                          size_type row_pitch = 0) const
        {
            /// Copy image slice to host memory.
            /// \param p pointer to host memory
            /// \param origin offset in device
            /// \param region size of image patch
            /// \param row_pitch pitch in host memory in pixels. Defaults to region[0]
            array<size_type, DIM> pitch;
            pitch[DIM - 1] = 0;
            if (DIM == 2)
                pitch[0] = row_pitch;

            buf.copy_to_host(p, flat_dim(imagenum, origin), flat_dim(1, region), pitch);
        }

        void copy_to_host(T* p) const
        {
            /// Copy image slice to host memory
            array<size_type, DIM> null;
            null.fill(0);
            copy_to_host(p, null, buf.dimensions());
        }

        template<typename EIF = void, typename = typename enable_if<!CONST, EIF>::type>
        void copy_from_host(const T* p, array<size_type, DIM> origin, array<size_type, DIM> region,
                            size_type row_pitch = 0)
        {
            /// copy image slice from host memory
            /// \param p pointer in host memory
            /// \param origin offset in device memory
            /// \param region size of image patch
            /// \param row_pitch pitch in host memory in pixels. Defaults to region[0]
            array<size_type, DIM> pitch;
            pitch[DIM - 1] = 0;
            if (DIM == 2)
                pitch[0] = row_pitch;

            buf.copy_from_host(p, flat_dim(imagenum, origin), flat_dim(1, region), pitch);
        }

        template<typename EIF = void, typename = typename enable_if<!CONST, EIF>::type>
        void copy_from_host(const T* p)
        {
            /// copy image slice from host memoryx
            array<size_type, DIM> null;
            null.fill(0);
            copy_from_host(p, null, buf.dimensions());
        }

        void fill(T value, array<size_type, DIM> origin, array<size_type, DIM> region)
        {
            buf.fill(value, flat_dim(imagenum, origin), flat_dim(1, region));
        }
        void fill(T value)
        {
            array<size_type, DIM> origin;
            origin.fill(0);
            fill(value, origin, dimensions());
        }
    };

    image_slice<false> operator[](size_type imagenum)
    {
        /// selects one image of the image array
        return { *this, imagenum };
    }
    image_slice<true> operator[](size_type imagenum) const
    {
        /// selects one image of the image array
        return { *this, imagenum };
    }


    inline void copy(const image_array_buffer& b,
                     array<size_type, DIM + 1> src_origin,
                     array<size_type, DIM + 1> dest_origin,
                     array<size_type, DIM + 1> region)
    {
        /// copy image array
        /// The last indices in src_origin, dest_origin, and region refer to the image slice, i.e., for 1D image arrays, coordinates in the form {x, imagenum} are passed,
        /// for 2D image arrays, coordinates in the form {x, y, imagenum} are passed.
        /// \param b source image array
        /// \param src_origin offset in source image array
        /// \param dest_origin offset in destination image array
        /// \param region size of image array patch.
        check_errno(interface::goopax_E_copy_image2(get_buf(), b.get_buf(), src_origin.data(), dest_origin.data(), region.data()));
    }

    inline void copy(const image_array_buffer& b)
    {
        /// copy image array
        if (dimensions() != b.dimensions())
        {
            exwhat() << "Different image dimensions while copying image.";
            throw EX::memory();
        }
        if (array_size() != b.array_size())
        {
            exwhat() << "Different array image size while copying image array.";
            throw EX::memory();
        }

        array<size_type, DIM + 1> null;
        null.fill(0);

        copy(b, null, null, flat_dim(array_size(), dimensions()));
    }

    void copy_to_host(T* p, array<size_type, DIM + 1> origin, array<size_type, DIM + 1> region,
                      array<size_type, DIM> pitchdim = makearray<size_type, DIM>(0)) const
    {
        /// copy image array to host memory.
        /// \param p pointer to host memory
        /// \param origin offset in device image
        /// \param region size of image patch
        /// \param pitchdim full image dimensions in host memory. If filled with zeros, will be calculated from image size
        check_errno(interface::goopax_E_copy_image_to_host2(this->get_buf(), p, origin.data(), region.data(), pitchdim.data()));
    }

    void copy_to_host(T* p) const
    {
        /// copy image array to host memory
        /// \param p pointer to host memory
        array<size_type, DIM + 1> origin;
        origin.fill(0);
        copy_to_host(p, origin, flat_dim(array_size(), dimensions()));
    }

    void copy_from_host(const T* p, array<size_type, DIM + 1> origin, array<size_type, DIM + 1> region,
                        array<size_type, DIM> pitchdim = makearray<size_type, DIM>(0))
    {
        /// copy image array from host memory
        /// \param p pointer in host memory
        /// \param origin image offset in device
        /// \param region size of image patch
        /// \param pitchdim full image dimensions in host memory. If filled with zeros, will be calculated from image size
        check_errno(interface::goopax_E_copy_image_from_host2(this->get_buf(), p, origin.data(), region.data(), pitchdim.data()));
    }

    void copy_from_host(const T* p)
    {
        // copy image array from host memory
        array<size_type, DIM + 1> origin;
        origin.fill(0);
        copy_from_host(p, origin, flat_dim(array_size(), dimensions()));
    }

    inline image_array_buffer()
    {
        /// default constructor. Image must not be used until replaced by a real image.
    }

    inline image_array_buffer(goopax_device               device,
                              size_type                   num_images,
                              const array<size_type, DIM> size,
                              IMAGE_FLAGS                 flags    = IMAGE_READ_WRITE,
                              value_type*                 host_ptr = nullptr,
                              array<size_type, DIM>       pitchdim = makearray<size_type, DIM>(0))
        : base_t(device, flat_dim(num_images, size), flags, host_ptr, pitchdim)
    {
        /**
           Allocated image array on the device.
           \param device: device to allocate image on
           \param num_devices number of images in the array
           \param size: image dimensions
           \param flags: must be one of IMAGE_READ_WRITE, IMAGE_READ_ONLY, IMAGE_WRITE_ONLY, IMAGE_WRITE_DISCARD
           \param host_ptr: if used, image array will be linked to the specified address in host memory.
           \param pitchdim: pitch size in pixels
           */
    }
    void fill(T value, array<size_type, DIM + 1> origin, array<size_type, DIM + 1> region)
    {
        /// FIll image array
        /// \param value color to fill
        /// \param origin origin in destination image. The last value specifies the image number in the array.
        /// \param region size of image patch to fill. The last value is the number of images to fill in the array.
        check_errno(interface::goopax_E_fill_image(this->get_buf(), &value, origin.data(), region.data()));
    }
    void fill(T value)
    {
        /// FIll image array
        /// \param value color to fill
        array<size_type, DIM + 1> origin;
        origin.fill(0);
        fill(value, origin, flat_dim(array_size(), dimensions()));
    }

    static inline image_array_buffer create_from_gl(goopax_device device,
                                                    Uint          GLres,
                                                    IMAGE_FLAGS   flags    = IMAGE_READ_WRITE,
                                                    Uint          GLtarget = (DIM == 1 ? 0x0DE0 : (DIM == 2 ? 0x0DE1 : 0x806F)) /* GL_TEXTURE_?D */,
                                                    Int           miplevel = 0);

    static inline image_array_buffer create_from_cl(goopax_device device, _cl_mem* mem);
};

/// \ingroup image
template<Uint DIM, Bool IS_ARRAY, typename T, Bool normalized>
struct image_resource_base : public image_type_info<T>
{
    static_assert(is_gpu_struct<T>::value, "");
    static_assert(!is_gpu_type<T>::value, "");
    using value_type = typename gettype<T>::gpu;

    using cpu_size_type  = Tuint;
    using cpu_ssize_type = typename make_signed<cpu_size_type>::type;
    using gpu_size_type  = typename gettype<cpu_size_type>::gpu;
    using gpu_ssize_type = typename gettype<cpu_ssize_type>::gpu;
    using size_type      = gpu_size_type;
    using ssize_type     = gpu_ssize_type;
    using access_type1   = typename goopax_struct_changetype<T, gpu_struct::struct_change_type<Tuint, typename image_type_info<T>::template channel_access_type<normalized>>>::type;
    using access_type    = typename goopax_struct_changetype<access_type1, typename get_X_type<gpu_uint>::type>::type;

    static const Uint num_dimensions = DIM;
    static const Bool is_checknan    = image_buffer<DIM, T, normalized>::is_checknan;
    static const Bool is_array       = IS_ARRAY;

    protected:
    static_assert(normalized ? image_type_info<T>::allow_normalized : image_type_info<T>::allow_unnormalized, "(un)normalized mode not allowed for this image type");
    inline size_type get_info(Tuint dim) const
    {
        size_type ret;
        interface::goopax_resource_info(this->impl.get(), ret.impl.get(), dim);
        return ret;
    }

    public:
    inline size_type width() const { return get_info(0); }
    template<typename EIF = void, typename = typename enable_if<(DIM >= 2), EIF>::type>
    inline size_type height() const
    {
        return get_info(1);
    }
    template<typename EIF = void, typename = typename enable_if<(DIM >= 3), EIF>::type>
    inline size_type depth() const
    {
        return get_info(2);
    }

    array<size_type, DIM> dimensions() const
    {
        /// image dimensions
        array<size_type, DIM> ret;
        for (Tuint k = 0; k < DIM; ++k)
            ret[k] = get_info(k);
        return ret;
    }


    unique_ptr<resource_impl, interface::goopax_deleter> impl;


    inline image_resource_base(image_resource_base&&) = default;
    inline image_resource_base& operator=(image_resource_base&&) = default;

    inline image_resource_base()
    {
        this->impl.reset(if_call(interface::goopax_E_image_resource_create2, DIM, (Bool)is_array, nullptr, false, this->channel_type_enum, this->num_channels, normalized));
    }
    inline image_resource_base(texbuf* buf)
    {
        this->impl.reset(if_call(interface::goopax_E_image_resource_create2, DIM, (Bool)is_array, buf, true, this->channel_type_enum, this->num_channels, normalized));
    }
    inline ~image_resource_base() = default;
};
template<Uint DIM, Bool IS_ARRAY, typename T, Bool normalized>
const Uint image_resource_base<DIM, IS_ARRAY, T, normalized>::num_dimensions;


/// \ingroup image
/// Image resources. Used for accessing an image_buffer from the kernel code.
/// \tparam DIM: Number of dimensions. Must be 1, 2, or 3.
/// \tparam T: Element data type. Must be std::array<C, N>, or other goopax struct of identical types. C: channel type (int32_t, int16_t, int8_t, uint32_t, uint16_t, uint8_t, float, half), N: number of channels (must be 4 at the moment).
/// \tparam normalized if set to true, image values are normalized to floats in the range 0..1 for unsigned channel types, or -1..1 for signed channel types. Defaults to false if the channel data type allows unnormalized access
template<Uint DIM, typename T, Bool normalized>
class image_resource : private image_resource_base<DIM, false, T, normalized>
{
    template<typename BUFREF, class R, typename IMAGE, class REST>
    friend struct ns_param_kernel::pas_common_image;
    using base_t = image_resource_base<DIM, false, T, normalized>;
    static_assert(DIM >= 1 && DIM <= 3, "");

    public:
    using buffer_type = image_buffer<DIM, T, normalized>;

    using typename base_t::access_type;
    using typename base_t::size_type;
    /// image dimensions
    using base_t::dimensions;
    using base_t::is_array;
    using base_t::is_checknan;
    using base_t::num_dimensions;

    using base_t::depth;
    using base_t::height;
    using base_t::width;
    /// true if samplers are allowed for the channel data type.
    static constexpr Bool allow_sampler = base_t::allow_sampler(normalized);

    private:
    template<typename AR>
    access_type read_impl(AR pos, sampler_mode sampler) const
    {
        using pos_t = typename gettype<typename AR::value_type>::gpu;
        static_assert(is_same<pos_t, gpu_float>::value || is_same<pos_t, gpu_int>::value || is_same<pos_t, gpu_uint>::value, "Please use float, int, or unsigned int for image read access!");

        access_type                                                    colors;
        array<const gpu_writer_base_impl*, DIM>                        pos2;
        array<gpu_writer_base_impl*, image_type_info<T>::num_channels> colors2;

        array<pos_t, DIM> tmp;
        for (Tuint k = 0; k < DIM; ++k)
        {
            tmp[k]  = pos[k];
            pos2[k] = tmp[k].impl.get();
        }
        for (Tuint k = 0; k < this->num_channels; ++k)
        {
            colors2[k] = colors[k].impl.get();
        }
        check_errno(interface::goopax_E_image_read3(this->impl.get(), pos2.data(), colors2.data(), sampler));
        return colors;
    }

    public:
    access_type read(array<size_type, DIM> pos) const
    {
        /// read from image
        /// \param pos image position
        /// \param return pixel at the given position
        return read_impl(pos, address_none | filter_nearest);
    }
    template<typename U, typename = typename enable_if<allow_sampler, U>::type>
    access_type read(array<U, DIM> pos, sampler_mode sampler) const
    {
        /// read from image with sampler
        /// \param pos position, may be an array of type gpu_int, gpu_uint, or gpu_float
        /// \param sampler mode.
        /// \param return pixel
        return read_impl(pos, sampler);
    }

    template<typename POS, typename = typename enable_if<is_same<typename POS::value_type, size_type>::value>::type>
    access_type read(POS pos) const
    {
        /// read pixel.
        /// \param pos pixel position. May be any container that provides the [] operator
        /// \param return pixel at the given position
        assert(pos.size() == DIM);
        return read_impl(pos, address_none | filter_nearest);
    }
    template<typename POS, typename = typename enable_if<allow_sampler, POS>::type>
    access_type read(POS pos, sampler_mode sampler) const
    {
        /// read from image with sampler
        /// \param pos position, may be any container that provides the [] operator, and that holds values of type gpu_int, gpu_uint, or gpu_float
        /// \param sampler mode.
        /// \param return pixel
        assert(pos.size() == DIM);
        return read_impl(pos, sampler);
    }


#ifdef _MSC_VER
    void write(array<size_type, DIM> pos, array<gpu_float, 4> colors)
#else
    void write(array<size_type, DIM> pos, access_type colors)
#endif
    {
        /// write to image
        /// \param pos pixel position
        /// \param colors pixel color to write
        array<const gpu_writer_base_impl*, DIM> pos2;
        array<const gpu_writer_base_impl*, 4>   colors2;
        for (Tuint k = 0; k < DIM; ++k)
        {
            pos2[k] = pos[k].impl.get();
        }
        for (Tuint k = 0; k < 4; ++k)
        {
            colors2[k] = colors[k].impl.get();
        }
        check_errno(interface::goopax_E_image_write2(this->impl.get(), pos2.data(), colors2.data()));
    }

#ifdef _MSC_VER
    template<typename COL, typename POS, typename = typename enable_if<is_same<typename POS::value_type, size_type>::value>::type>
    void write(POS pos, COL colors)
#else
    template<typename POS, typename = typename enable_if<is_same<typename POS::value_type, size_type>::value>::type>
    void write(POS pos, access_type colors)
#endif
    {
        /// write to image
        /// \param pos pixel position
        /// \param colors pixel color to write. May be any container that provides the [] operator
        assert(pos.size() == DIM);
        array<size_type, DIM> pos2;
        for (Tuint k = 0; k < DIM; ++k)
            pos2[k] = pos[k];
        write(pos2, colors);
    }


    inline image_resource(image_resource&&) = default;
    inline image_resource& operator=(image_resource&&) = default;

    private:
    inline image_resource() {}

    public:
    inline image_resource(buffer_type& tex)
        : base_t(tex.get_buf())
    {
        /// constructor.
        /// \param tex image_buffer to link to. The template parameters of the image_buffer and the image_resource must match
    }
    inline ~image_resource() = default;
};


/// \ingroup image
/// Image array resources. Used for accessing an image_array_buffer from the kernel code.
/// \tparam DIM: Number of dimensions. Must be 1 or 2.
/// \tparam T: Element data type. Must be std::array<C, N>, or other goopax struct of identical types. C: channel type (int32_t, int16_t, int8_t, uint32_t, uint16_t, uint8_t, float, half), N: number of channels (must be 4 at the moment).
/// \tparam normalized if set to true, image values are normalized to floats in the range 0..1 for unsigned channel types, or -1..1 for signed channel types. Defaults to false if the channel data type allows unnormalized access
template<Uint DIM, typename T, Bool normalized>
class image_array_resource : private image_resource_base<DIM, true, T, normalized>
{
    static_assert(DIM == 1 || DIM == 2, "Only 1D and 2D image arrays are supported.");
    template<typename BUFREF, class R, typename IMAGE, class REST>
    friend struct ns_param_kernel::pas_common_image;

    private:
    using base_t = image_resource_base<DIM, true, T, normalized>;
    static_assert(DIM >= 1 && DIM <= 3, "");

    public:
    using buffer_type = image_array_buffer<DIM, T, normalized>;

    using base_t::num_dimensions;
    using typename base_t::access_type;
    using typename base_t::size_type;
    /// image dimensions
    using base_t::depth;
    using base_t::dimensions;
    using base_t::height;
    using base_t::is_array;
    using base_t::is_checknan;
    using base_t::width;
    static constexpr Bool allow_sampler = base_t::allow_sampler(normalized);

    inline size_type array_size() const { return this->get_info(DIM); }

    /// Reference to single image in the image array
    template<typename RESREF>
    struct image_slice
    {
        RESREF    buf;
        size_type imagenum;

        inline size_type width() const
        {
            return buf.width();
        }
        template<typename EIF = void, typename = typename enable_if<(DIM >= 2), EIF>::type>
        inline size_type height() const
        {
            return buf.height();
        }


        private:
        template<typename AR>
        access_type read_impl(AR pos, sampler_mode sampler) const
        {
            using pos_t = typename gettype<typename AR::value_type>::gpu;
            static_assert(is_same<pos_t, gpu_float>::value || is_same<pos_t, gpu_int>::value || is_same<pos_t, gpu_uint>::value, "Please use float, int, or unsigned int for image read access!");

            access_type                                                    colors;
            array<const gpu_writer_base_impl*, DIM + 1>                    pos2;
            array<gpu_writer_base_impl*, image_type_info<T>::num_channels> colors2;

            array<pos_t, DIM + 1> tmp;

            for (Tuint k = 0; k < DIM + 1; ++k)
            {
                tmp[k]  = (k == DIM ? imagenum : pos[k]);
                pos2[k] = tmp[k].impl.get();
            }
            for (Tuint k = 0; k < buf.num_channels; ++k)
            {
                colors2[k] = colors[k].impl.get();
            }
            check_errno(interface::goopax_E_image_read3(buf.impl.get(), pos2.data(), colors2.data(), sampler));
            return colors;
        }

        public:
        access_type read(array<size_type, DIM> pos) const
        {
            /// read from image slice
            /// \param pos image position
            /// \param return pixel at the given position
            return read_impl(pos, address_none | filter_nearest);
        }
        template<typename U, typename = typename enable_if<allow_sampler, U>::type>
        access_type read(array<U, DIM> pos, sampler_mode sampler = address_none | filter_nearest) const
        {
            /// read from image slice with sampler
            /// \param pos position, may be an array of type gpu_int, gpu_uint, or gpu_float
            /// \param sampler mode.
            /// \param return pixel
            return read_impl(pos, sampler);
        }
        template<typename POS, typename = typename enable_if<is_same<typename POS::value_type, size_type>::value>::type>
        access_type read(POS pos) const
        {
            /// read pixel.
            /// \param pos pixel position. May be any container that provides the [] operator
            /// \param return pixel at the given position
            assert(pos.size() == DIM);
            return read_impl(pos, address_none | filter_nearest);
        }
        template<typename POS, typename = typename enable_if<allow_sampler, POS>::type>
        access_type read(POS pos, sampler_mode sampler) const
        {
            /// read from image with sampler
            /// \param pos position, may be any container that provides the [] operator, and that holds values of type gpu_int, gpu_uint, or gpu_float
            /// \param sampler mode.
            /// \param return pixel
            assert(pos.size() == DIM);
            return read_impl(pos, sampler);
        }


        template<typename = void, typename = typename enable_if<is_same<RESREF, typename remove_const<RESREF>::type>::value>::type>
        void write(array<size_type, DIM> pos, access_type colors)
        {
            /// write to image
            /// \param pos pixel position
            /// \param colors pixel color to write
            array<const gpu_writer_base_impl*, DIM + 1>                          pos2;
            array<const gpu_writer_base_impl*, image_type_info<T>::num_channels> colors2;

            array<size_type, DIM + 1> tmp;

            for (Tuint k = 0; k < DIM + 1; ++k)
            {
                tmp[k]  = (k == DIM ? imagenum : pos[k]);
                pos2[k] = tmp[k].impl.get();
            }
            for (Tuint k = 0; k < image_type_info<T>::num_channels; ++k)
            {
                colors2[k] = colors[k].impl.get();
            }
            check_errno(interface::goopax_E_image_write2(buf.impl.get(), pos2.data(), colors2.data()));
        }

        template<typename POS, typename = typename enable_if<is_same<typename POS::value_type, size_type>::value>::type>
        void write(POS pos, access_type colors)
        {
            /// write to image
            /// \param pos pixel position
            /// \param colors pixel color to write. May be any container that provides the [] operator
            assert(pos.size() == DIM);
            array<size_type, DIM> pos2;
            for (Tuint k = 0; k < DIM; ++k)
                pos2[k] = pos[k];
            write(pos2, colors);
        }
    };

    image_slice<image_array_resource&> operator[](size_type imagenum)
    {
        /// access image slice
        /// \param imagenum image number in the array
        /// \return image slice
        return { *this, imagenum };
    }
    image_slice<const image_array_resource&> operator[](size_type imagenum) const
    {
        /// access image slice (const)
        /// \param imagenum image number in the array
        /// \return image slice
        return { *this, imagenum };
    }


    inline image_array_resource(image_array_resource&&) = default;
    inline image_array_resource& operator=(image_array_resource&&) = default;

    private:
    inline image_array_resource() {}

    public:
    inline image_array_resource(const buffer_type& tex)
        : base_t(tex.get_buf())
    {
        /// constructor.
        /// \param tex image_array_buffer to link to. The template parameters of the image_array_buffer and the image_array_resource must match
    }
    inline ~image_array_resource() = default;
};


}
using impl::image_array_buffer;
using impl::image_array_resource;
using impl::image_buffer;
using impl::image_resource;

using impl::const_image_array_buffer_map;
using impl::const_image_buffer_map;
using impl::image_array_buffer_map;
using impl::image_buffer_map;

}

/// \}
