// -*- C++ -*-

#pragma once

#include "memory.h"

/**
\addtogroup loops
Loops in gpu kernels can be achieved by the use of gpu_for or gpu_while statements.
Alternatively, small loops that don't depend on gpu variables may also be implemented as standard C++ for/while loops. This will lead to explicit unrolling of the loop.

The usage of the gpu_for loop is different from C++ for loops and looks like this:
\code{.cpp}
gpu_for(gpu_uint k=0, 10)
{
  <do something> // k runs from 0 to 9
}
\endcode

Step sizes can be provided as a third argumend:
 \code{.cpp}
 gpu_for(gpu_int k=0, 10, 2)
 {
    <do something> // only runs over even numbers form 0 to 8
 }
 \endcode
 
 The comparison operator can be specified as a fourth argument as a C string (defaults to "<"):
 \code{.cpp}
 gpu_for(gpu_int k=0, 10, 2, "<=")
 {
    <do something> // only runs over even numbers form 0 to 10
 }
 \endcode

\{
*/


namespace goopax {
namespace impl {
namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(3.0.0) gpu_if_impl* goopax_if_create2(const char*, Uint, gpu_writer_base_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) gpu_if_impl* goopax_elseif_create2(const char*, Uint, gpu_writer_base_impl*) noexcept;
GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_loop_break(loop_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) void goopax_infloop_set_next2(loop_impl*, gpu_writer_base_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) Bool goopax_infloop_do_continue2(loop_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) loop_impl* goopax_infloop_create3(const char*, Uint, gpu_writer_base_impl*) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) void* goopax_loop_access2() noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_loop_make2(void*, const char*, Uint, Uint, const gpu_writer_base_impl*, const gpu_writer_base_impl*, const char*) noexcept;
}
}

template<typename P, memory::scope scope>
class gpu_ptrbase;


struct gpu_if_clause
{
    goopax::impl::Tuint                                usecount = 0;
    unique_ptr<gpu_if_impl, interface::goopax_deleter> impl;

    inline gpu_if_clause(const char* FILE, Uint LINE, const gpu_bool& cnd)
        : impl(interface::goopax_if_create2(FILE, LINE, cnd.impl.get()))
    {}
    inline gpu_if_clause(const char* FILE, Uint LINE, Tint, const gpu_bool& cnd)
        : impl(interface::goopax_elseif_create2(FILE, LINE, cnd.impl.get()))
    {}
    gpu_if_clause(const gpu_if_clause&) = delete;
    gpu_if_clause& operator=(const gpu_if_clause&) = delete;
};


inline void gpu_break()
{
    check_errno(interface::goopax_E_loop_break3(nullptr));
}
inline void gpu_continue()
{
    check_errno(interface::goopax_E_loop_continue(nullptr));
}


struct infinite_loop
{
    protected:
    unique_ptr<loop_impl, interface::goopax_deleter> impl;

    public:
    inline void loop_set_next(const gpu_bool& cont)
    {
        interface::goopax_infloop_set_next2(impl.get(), cont.impl.get());
    }
    inline goopax::impl::Tbool loop_do_continue() const
    {
        return interface::goopax_infloop_do_continue2(impl.get());
    }

    inline infinite_loop(const char* FILE, Uint LINE, const gpu_bool& initial)
        : impl(interface::goopax_infloop_create3(FILE, LINE, initial.impl.get()))
    {}

    inline ~infinite_loop() = default;
};


/// \ingroup loops
/// gpu_if, to be used in kernels. Takes gpu_bool as input. The block following the gpu_if() statement is executed by all threads for which the argument is true.
#define gpu_if(...) for (goopax::impl::gpu_if_clause __goopax_if_obj(__FILE__, __LINE__, static_cast<goopax::gpu_bool>(__VA_ARGS__)); __goopax_if_obj.usecount == 0; ++__goopax_if_obj.usecount)
/// \ingroup loops
/// gpu_elseif, must follow a gpu_if or gpu_elseif block. The block following the gpu_elseif() statement is executed by all threads for which the argument is true.
#define gpu_elseif(...) for (goopax::impl::gpu_if_clause __goopax_if_obj(__FILE__, __LINE__, 1, static_cast<goopax::gpu_bool>(__VA_ARGS__)); __goopax_if_obj.usecount == 0; ++__goopax_if_obj.usecount)
/// \ingroup loops
/// gpu_elseif must follow a gpu_if or gpu_elseif block. The block following the gpu_elseif statement is executed by all threads that did not execute the previous if blocks.
#define gpu_else gpu_elseif(true)

/// \ingroup loops
/// while loop.
#define gpu_while(...) for (goopax::impl::infinite_loop LOOP(__FILE__, __LINE__, __VA_ARGS__); LOOP.loop_do_continue(); LOOP.loop_set_next(__VA_ARGS__))


template<typename T>
gpu_type<T> get_loop_type(gpu_type<T>, ...);


struct loop_ptr_converter
{
    void* ptr;
    template<typename T>
    operator T*()
    {
        return static_cast<T*>(ptr);
    }
};

template<typename T>
static T* loop_make(T* var, const char* FILE, Uint LINE, Tuint MODE, typename enable_if<true, T>::type end, typename enable_if<true, T>::type step = 1, const char* cmp = "<")
{
    check_errno(interface::goopax_E_loop_make2(var, FILE, LINE, MODE, end.impl.get(), step.impl.get(), cmp));
    return var;
}

template<typename T, memory::scope scope>
static gpu_type<T*, scope>* loop_make(gpu_type<T*, scope>* var, const char* FILE, Uint LINE, Tuint MODE, typename enable_if<true, gpu_type<T*, scope>>::type end, gpu_int64 step = 1, const char* cmp = "<")
{
    constexpr Size_t origsize = sizeof(typename goopax::goopax_struct_changetype<typename remove_cv<T>::type, Uint>::type);
    check_errno(interface::goopax_E_loop_make2(var, FILE, LINE, MODE, end.impl.get(), (step * (origsize / gettype<typename gpu_ptrbase<T*, scope>::basetype>::size)).impl.get(), cmp));
    //	check_errno(interface::goopax_E_loop_make2(var, FILE, LINE, MODE, end.impl.get(), (step * (gettype<T>::size / gettype<typename gpu_ptrbase<T*, T>::basetype>::size)).impl.get(), cmp));
    return var;
}

#ifdef _MSC_VER
#define gpu_for(a, ...) for (a + 0, *GOOPAX_loopvar_access = goopax::impl::loop_ptr_converter({ goopax::impl::interface::goopax_loop_access2() }), *GOOPAX_loopvar_assign = goopax::impl::loop_make(GOOPAX_loopvar_access, __FILE__, __LINE__, 0, __VA_ARGS__); GOOPAX_loopvar_assign != nullptr; GOOPAX_loopvar_assign = nullptr)
#define gpu_for_local(a, ...) for (a + 0, *GOOPAX_loopvar_access = goopax::impl::loop_ptr_converter({ goopax::impl::interface::goopax_loop_access2() }), *GOOPAX_loopvar_assign = goopax::impl::loop_make(GOOPAX_loopvar_access, __FILE__, __LINE__, 1, __VA_ARGS__); GOOPAX_loopvar_assign != nullptr; GOOPAX_loopvar_assign = nullptr)
#define gpu_for_global(a, ...) for (a + 0, *GOOPAX_loopvar_access = goopax::impl::loop_ptr_converter({ goopax::impl::interface::goopax_loop_access2() }), *GOOPAX_loopvar_assign = goopax::impl::loop_make(GOOPAX_loopvar_access, __FILE__, __LINE__, 2, __VA_ARGS__); GOOPAX_loopvar_assign != nullptr; GOOPAX_loopvar_assign = nullptr)
#define gpu_for_group(a, ...) for (a + 0, *GOOPAX_loopvar_access = goopax::impl::loop_ptr_converter({ goopax::impl::interface::goopax_loop_access2() }), *GOOPAX_loopvar_assign = goopax::impl::loop_make(GOOPAX_loopvar_access, __FILE__, __LINE__, 3, __VA_ARGS__); GOOPAX_loopvar_assign != nullptr; GOOPAX_loopvar_assign = nullptr)
#else

#define goopax_gpu_for(MODE, a, ...) for (a, *GOOPAX_loopvar_access = goopax::impl::loop_ptr_converter({ goopax::impl::interface::goopax_loop_access2() }), *GOOPAX_loopvar_assign = goopax::impl::loop_make(GOOPAX_loopvar_access, __FILE__, __LINE__, MODE, __VA_ARGS__); GOOPAX_loopvar_assign != nullptr; GOOPAX_loopvar_assign = nullptr)


/// \brief gpu for loop.
/// \ingroup loops
/// usage:
/// gpu_for(\<GPU TYPE\> \<var\>=\<start\>, \<end\> [,\<step\>])
/// {
/// ...
/// }
#define gpu_for(...) goopax_gpu_for(0, __VA_ARGS__)
/// \ingroup loops
/// gpu loop, parallelized over threads in workgroup.
#define gpu_for_local(...) goopax_gpu_for(1, __VA_ARGS__)
/// \ingroup loops
/// gpu loop, parallelized over all threads
#define gpu_for_global(...) goopax_gpu_for(2, __VA_ARGS__)
/// \ingroup loops
/// gpu loop, parallelized over workgroups
#define gpu_for_group(...) goopax_gpu_for(3, __VA_ARGS__)
#endif

}
using impl::gpu_break;
using impl::gpu_continue;
}

/// \}
