// -*- C++ -*-

#pragma once

#include "gpu-utils.h"
#include "il_writer_utils.h"
#include "kernel.h"
#include "memory.h"
#include "pointer.h"
#include "struct.h"
#include <unordered_map>

/// \addtogroup random random numbers
/// \{
/// Random numbers can be generated with the WELL512 random number generator. The current implementation is optimized for high thoughput, but
/// it has the limitation that random numbers are generated in 512 bit blocks.
/// To use the random number generator, a pair of classes need to be instantiated: The WELL512 class in host code and the WELL512lib class in device code.
/// The WELL512 class is used for storing the random seed between kernel calls, the WELL512lib class provides the actual generator of the random numbers.
/// See the pi example program for an example implementation.

namespace goopax {
namespace impl {
class WELL512;

namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_well512_create2(WELL512_impl**, goopax_device_impl* device, WELL512* deleteptr, void (*deletefunc0)(WELL512*, goopax_device_impl*)) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_well512_setseed(WELL512_impl*, const Tuint32_t*, Uint32_t);
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_rnd_vecwell_create2(vecWELL512lib_impl**, WELL512_impl*) noexcept;

GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_well512_genvec2(Uint*, vecWELL512lib_impl*, Uint) noexcept;
GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_well512_genvec_extract2(vecWELL512lib_impl*, gpu_writer_base_impl*) noexcept;
}

}


/// \ingroup random
/// Random number generator, based on the WELL512 algorithm.
/// An object of this class must be instantiated to be used from kernels.
class WELL512
{
    std::unordered_map<goopax_device_impl*, unique_ptr<WELL512_impl, interface::goopax_deleter>> impl;
    friend class WELL512lib;

    WELL512_impl* get_impl(goopax_device_impl* device)
    {
        auto p = impl.find(device);
        if (p == impl.end())
        {
            impl.emplace(device, unique_ptr<WELL512_impl, interface::goopax_deleter>(if_call(interface::goopax_E_well512_create2, device, this, deletefunc)));
            p = impl.find(device);
        }
        return p->second.get();
    }
    static void deletefunc(WELL512* THIS, goopax_device_impl* device)
    {
        THIS->impl.erase(device);
    }

    public:
    void seed(goopax_device device, const vector<Tuint32_t>& seed)
    {
        /// \brief Set random seed.
        /// All threads will get an additional offset by goopax, so that they all produce different random numbers.
        /// \param device device for which to set the random seed
        /// \param seed seed data, may be any size, maximum 16.
        check_errno(interface::goopax_E_well512_setseed(get_impl(device.get_impl()), seed.data(), seed.size()));
    }
};

/// \ingroup random
/// Random number generator, used in device code
/// This object must be linked to an WELL512 object in host code
class WELL512lib
{
    public:
    unique_ptr<vecWELL512lib_impl, interface::goopax_deleter> impl;

    template<typename T>
    vector<typename gettype<T>::gpu> gen_vec()
    {
        /// Generate uniformly distributed random numbers of type T. With each call, a vector of values is created. For optimal performance, all values in the vector should be used.
        /// Floating point values are normalized in the range [0,1). Integer values are equally distributed over all possibly values.
        static_assert(!is_same<typename gettype<T>::cpu, Tbool>::value, "random vector of booleans not supported.");
        vector<typename gettype<T>::gpu> ret(if_call(interface::goopax_E_well512_genvec2, impl.get(), type_enum<typename gettype<T>::cpu>::value));
        for (Tsize_t k = 0; k < ret.size(); ++k)
        {
            check_errno(interface::goopax_E_well512_genvec_extract2(impl.get(), ret[k].impl.get()));
        }
        return ret;
    }

    WELL512lib(WELL512& well)
        : impl(if_call(interface::goopax_E_rnd_vecwell_create2, well.get_impl(interface::goopax_build_get_current_device())))
    {
        /// Constructor. Must be linked to an existing WELL512 object.
        /// \param well existing WELL512 object.
    }
    WELL512lib(const WELL512lib&) = delete;
    WELL512lib& operator=(const WELL512lib&) = delete;
};
}


using impl::WELL512;
using impl::WELL512lib;
}
/// \}
